(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ceta_Checker
imports Ceta
begin

subsection {* User exceptions in generated code *}
(* FIXME:  will be moved to Main *)

definition error :: "String.literal \<Rightarrow> 'a \<Rightarrow> 'a" 
where [code del]: "error msg x = x"

code_const error
  (Eval "(fn msg =>/ fn x =>/ error msg)")

code_reserved Eval error

subsection {* Checker *}

text {* Assemble a checker with a simple interface from the bits and pieces provided by IsaFoR. *}

(* FIXME: Shouldn't the errors be uniform (string vs. show)??? *)
fun release_shows_error :: "'a \<Rightarrow> shows + 'a \<Rightarrow> 'a"
where
  "release_shows_error d (Inl msg) = error (implode (msg [])) d"
| "release_shows_error d (Inr x) = x"

fun release_string_error :: "'a \<Rightarrow> 'a sparse_result \<Rightarrow> 'a"
where
  "release_string_error d (PError msg) = error (implode msg) d"
| "release_string_error d (PResult x) = x"

lemma [code]: "release_string_error d (parseresult (Inl msg)) = error (implode msg) d" by simp
lemma [code]: "release_string_error d (parseresult (Inr x)) = x" by simp

definition ceta_checker :: "(string, string) rules \<Rightarrow>
  (string, label_type, string) trs_proof \<Rightarrow> bool"
where
  "ceta_checker trs prf = 
    release_shows_error False 
      (check_trs_proof False
        (shows_string ''1'')
        tp_impl
        dpp_impl
        (tp_ops.mk tp_impl [] (init_lab_rules trs) [])
        prf
      \<guillemotright>= (\<lambda>x. Error_Monad.return True))"

lemma ceta_checker:
  assumes check: "ceta_checker trs prf"
  shows "SN (rstep (set trs))"
proof -
  let ?Q = "[]::((string, label_type) lab, string) term list"
  let ?S = "[]::((string, label_type) lab, string) rules"
  let ?R = "init_lab_rules trs::((string, label_type) lab, string) rules"
  let ?RR = "init_lab_trs (set trs)::((string, label_type) lab, string) trs"
  let ?call = "check_trs_proof False
    (shows_string ''1'') tp_impl dpp_impl (tp_ops.mk tp_impl ?Q ?R ?S) prf"
  show ?thesis 
  proof (cases ?call)
    case Inl with check show ?thesis by (auto simp: error_def ceta_checker_def)
  next
    case Inr
    then have "isOK (?call)" by simp
    from check_trs_proof_sound[OF tp_impl dpp_impl this]
    have "SN (rstep (set ?R))"
      unfolding tp_spec.mk_sound[OF tp_impl] SN_qrel_def split
      by simp
    hence SN: "SN (rstep ?RR)"
      unfolding map_trs.simps by auto
    {
      fix t
      have "SN_on (rstep (set trs)) {t}"
        by (rule SN_map_imp_SN[OF SN[unfolded SN_def, THEN spec]])
    }
    from SN_I[OF this] show ?thesis .
  qed
qed

definition dummy_proof :: "(string, label_type, string) trs_proof" where
  "dummy_proof = RisEmpty"

definition parse_trs_proof :: "string \<Rightarrow> Code_Evaluation.term" where
  "parse_trs_proof s = Code_Evaluation.term_of 
  (case release_string_error
    (TRS_Proof No_Strategy [] None dummy_proof) (parse_cert_problem s) of
    TRS_Proof _ _ _ prf \<Rightarrow> prf
  | _ \<Rightarrow> error (STR ''not a termination proof'') dummy_proof)"

ML {*
structure Ceta_Parser =
struct
  val read_cpf = 
    String.explode
    #> @{code parse_trs_proof}
    #> Pattern.rewrite_term @{theory} 
      [(term_of @{cpat "Frct (?x, ?y)"}, term_of @{cpat "Fract ?x ?y"})] []
end
*}

end
