(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_DPP
imports Check_Common
begin

subsection {*Proving finiteness in the QDP Framework*}

definition missing :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "missing s x \<equiv> ''the '' +#+ s +#+ '' '' +#+ x +@+ shows_string '' is missing''"

definition toomuch :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "toomuch s x \<equiv> ''superfluous '' +#+ s +#+ '' '' +#+ x +@+ x"

fun
  check_dpp_subsumes ::
    "('d, ('f::show, 'l::show) lab, 'v::show) dpp_ops
    \<Rightarrow> ('f, 'l, 'v) dppLL \<Rightarrow> 'd \<Rightarrow> shows check"
where
  "check_dpp_subsumes I (p', pw', q', r', rw') d = (do {
    let p  = dpp_ops.P I d;
    let pw = dpp_ops.Pw I d;
    let q  = dpp_ops.Q I d;
    let r  = dpp_ops.R I d;
    let rw = dpp_ops.Rw I d;
    let pb' = p' @ pw';
    let rb  = r  @ rw;
    let rb' = r' @ rw';
    check_subseteq p p'   <+? (\<lambda>r. toomuch ''pair'' (shows_rule r));
    check_subseteq pw pb' <+? (\<lambda>r. toomuch ''weak pair'' (shows_rule r));
    check_NF_terms_eq q q'   <+? (\<lambda>r. shows ''NF(Q) differs due to term '' +@+ shows_term r);
    check_subseteq r r'   <+? (\<lambda>r. toomuch ''strict rule'' (shows_rule r));
    check_subseteq rb rb' <+? (\<lambda>r. toomuch ''strict/weak rule'' (shows_rule r));
    check_subseteq rb' rb <+? (\<lambda>r. missing ''strict/weak rule'' (shows_rule r));
    return ()
  }) <+? (\<lambda>s. ''finiteness of the problem'' +#+ shows_nl +@+ shows_dpp I d +@+ shows_nl
           +@+ ''may not be concluded from assuming finiteness of the problem'' +#+ shows_nl
           +@+ shows_dpp I (dpp_ops.mk I p' pw' q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

lemma check_dpp_subsumesE[elim]:
  assumes 1: "dpp_spec I"
    and 2: "isOK (check_dpp_subsumes I (p, pw, q, r, rw) d)"
    and 3: "set (dpp_ops.P I d) \<subseteq> set p
    \<Longrightarrow> set (dpp_ops.Pw I d) \<subseteq> set p \<union> set pw
    \<Longrightarrow> NF_terms (set (dpp_ops.Q I d)) = NF_terms (set q)
    \<Longrightarrow> set (dpp_ops.R I d) \<subseteq> set r
    \<Longrightarrow> set (dpp_ops.R I d) \<union> set (dpp_ops.Rw I d) = set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2, auto simp: Let_def)

fun finite_dppL :: "('f, 'l, 'v) dppLL \<Rightarrow> bool" where
  "finite_dppL (p, pw, q, r, rw) = finite_dpp (set p, set pw, set q, set r, set rw)"

lemma check_dpp_subsumes_sound:
  assumes I: "dpp_spec I"
    and ok: "isOK (check_dpp_subsumes I dpp' dpp)"
    and fin: "finite_dppL dpp'"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
proof -
  obtain p pw q r rw where dpp': "dpp' = (p, pw, q, r, rw)" by (cases dpp') force+
  let ?P   = "set (dpp_ops.P I dpp)"
  let ?Pw  = "set (dpp_ops.Pw I dpp)"
  let ?Q   = "set (dpp_ops.Q I dpp)"
  let ?R   = "set (dpp_ops.R I dpp)"
  let ?Rw  = "set (dpp_ops.Rw I dpp)"
  let ?P'  = "set p"
  let ?Pw' = "set pw"
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"
  from I and ok
    have "?P \<subseteq> ?P'" and "?Pw \<subseteq> ?P' \<union> ?Pw'" and q: "NF_terms ?Q = NF_terms ?Q'" and "?R \<subseteq> ?R'"
    and "?R \<union> ?Rw = ?R' \<union> ?Rw'"
    unfolding dpp' by blast+
  note 1 = this
  show ?thesis unfolding dpp_spec.dpp_sound[OF I]
    by (rule finite_dpp_mono[OF fin[unfolded dpp' finite_dppL.simps]], insert 1, auto)
qed

datatype ('f, 'l, 'v) dp_proof =
  PisEmpty
| SubtermCriterionProc "('f, 'l) lab projL" "('f, 'l, 'v) rseqL"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| RedPairProc "('f,'l)lab root_redtriple_impl + ('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL"  "('f, 'l, 'v) dp_proof"
| RedPairUrProc "('f, 'l) lab root_redtriple_impl + ('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| UsableRulesProc "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| DepGraphProc "(('f, 'l, 'v) dp_proof option \<times> ('f, 'l, 'v) trsLL) list"
| MonoRedPairProc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof" 
| MonoRedPairUrProc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| SizeChangeSubterm "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| SizeChangeRedPair "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL option"
    "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| UncurryProc "nat option" "(('f, 'l) lab, 'v) uncurry_info"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| FccProc "('f, 'l) lab" "(('f, 'l) lab, 'v) ctxt list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_proof"
| SplitProc
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" 
    "('f, 'l, 'v) dp_proof" "('f, 'l, 'v) dp_proof"
| SemlabProc
    "(('f, 'l) lab, 'v) sl_variant" "('f, 'l, 'v) trsLL"
    "(('f, 'l) lab, 'v) term list" "('f, 'l, 'v) trsLL"
    "('f, 'l, 'v) dp_proof"
| SwitchInnermostProc "('f,'l)lab join_info" "('f, 'l, 'v)dp_proof"
| RewritingProc   "('f, 'l, 'v) trsLL" "('f,'l,'v)ruleLL" "('f,'l,'v)ruleLL" "('f,'l,'v)ruleLL" pos "('f, 'l, 'v) dp_proof"
| InstantiationProc   "('f,'l,'v)ruleLL" "('f,'l,'v)trsLL" "('f, 'l, 'v) dp_proof"
| ForwardInstantiationProc   "('f,'l,'v)ruleLL" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL option" "('f, 'l, 'v) dp_proof"
| NarrowingProc   "('f,'l,'v)ruleLL" pos "('f,'l,'v)trsLL" "('f, 'l, 'v) dp_proof"
| AssumeFinite "('f, 'l, 'v) dppLL"
| UnlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)dp_proof"
| QReductionProc "('f,'l,'v)termsLL" "('f,'l,'v)dp_proof"


fun get_fcc_option :: "('f,'l,'v)dp_proof \<Rightarrow> (('f, 'l) lab \<times> (('f, 'l) lab, 'v) ctxt list \<times> ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) dp_proof)option"
  where "get_fcc_option (FccProc f fcs Pb' Rb' prf) = Some (f,fcs,Pb',Rb',prf)"
     |  "get_fcc_option _ = None"

(*The first argument is a flag that specifies whether partial proofs (i.e, containing
assumptions) are accepted or not.*)
function 
  check_dp_proof ::
    "bool \<Rightarrow> shows \<Rightarrow> ('dpp, ('f::{show,linorder},label_type) lab, string) dpp_ops \<Rightarrow> 'dpp \<Rightarrow>
    ('f, label_type, string) dp_proof \<Rightarrow>
    shows check"
where 
  "check_dp_proof assms i I dpp PisEmpty = debug (i []) ''P is empty''
     (if dpp_ops.P I dpp = [] \<and> (dpp_ops.Pw I dpp = [] \<or> dpp_ops.R I dpp = []) 
          then return () 
          else error (i +@+ '': P is not empty in the following DP-problem '' +#+
         shows_nl +@+ shows_dpp I dpp))"
| "check_dp_proof assms i I dpp (SubtermCriterionProc p rseq rP prf) =
    debug (i []) ''SubtermCriterionProc'' (do {
      let P  = dpp_ops.pairs I dpp;
      let Pr = list_diff P rP;
      dpp' \<leftarrow> subterm_criterion_proc I (create_proj p) (create_rseq_map rseq) Pr dpp
        <+? (\<lambda>s. i +@+ '': error when applying the subterm criterion to the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
          ''and trying to remove the pairs '' +#+ shows_nl +@+ shows_rules Pr
          +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the subterm criterion'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (RedPairProc redp rP prf) = debug (i []) ''RedPairProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> generic_ur_af_redtriple_proc I (get_redtriple rp) None
                        | Inl rrp \<Rightarrow> generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) None);
         dpp' \<leftarrow> Proc Pr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (UsableRulesProc U prf) = debug (i []) ''UsableRulesProc'' (do {
         dpp' \<leftarrow> usable_rules_proc I U dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the usable rules processor to restrict the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' to the usable rules '' +#+ shows_nl +@+ shows_rules U +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the usable rules processor'' +#+ shows_nl
             +@+ indent s)
   })"
| "check_dp_proof assms i I dpp (QReductionProc Q prf) = debug (i []) ''QReductionProc'' (do {
         dpp' \<leftarrow> q_reduction_proc I Q dpp 
           <+? (\<lambda>s. i +@+ shows '': error when applying the Q-reduction processor'' +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the Q-reduction processor'' +#+ shows_nl
             +@+ indent s)
   })"
|  "check_dp_proof assms i I dpp (MonoRedPairProc redp rP rR prf) = debug (i []) ''MonoRedPairProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = list_diff R rR;
         dpp' \<leftarrow> mono_redpair_proc I (get_redtriple redp) Pr Rr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (DepGraphProc edpts) =
     debug (i []) ''DepGraphProc'' (do {
        pdpps \<leftarrow> dep_graph_proc I dpp edpts
         <+? (\<lambda>s. i +@+ '': error while trying to perform SCC-decomposition  on '' +#+
           shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
        check_allm_index (\<lambda> (prof,dpp') j. check_dp_proof assms (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) pdpps
          <+? (\<lambda>s. i +@+ '': error below the dependency graph processor'' +#+ shows_nl
            +@+ indent s)
  })"
| "check_dp_proof assms i I dpp (RedPairUrProc redp rP ur prf) = debug (i []) ''RedPairUrProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> generic_ur_af_redtriple_proc I (get_redtriple rp) (Some ur) 
                        | Inl rrp \<Rightarrow> generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) (Some ur));
         dpp' \<leftarrow> Proc Pr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_dp_proof assms i I dpp (MonoRedPairUrProc redp rP rR ur prf) = debug (i []) ''MonoRedPairUrProc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = list_diff R rR;
         dpp' \<leftarrow> mono_ur_redpair_proc I (get_redtriple redp) Pr Rr ur dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor with usable rules'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (UncurryProc mode u_info P' R' prf) = 
    debug (i []) ''UncurryProc'' (do {
      dpp' \<leftarrow> uncurry_proc_both I mode u_info P' R' dpp
        <+? (\<lambda>s. i +@+ '': error when applying the uncurrying processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the uncurrying processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (SizeChangeSubterm graphs) =
     debug (i []) ''SizeChangeSubterm'' (sct_subterm_proc I graphs dpp)"
| "check_dp_proof assms i I dpp (SizeChangeRedPair redp U_opt graphs) =
     debug (i []) ''SizeChangeRedPair'' (sct_ur_af_proc I (get_redtriple redp) graphs U_opt dpp)"
| "check_dp_proof assms i I dpp (FccProc f fcs Pb' R' prf) = debug (i []) ''FccProc'' (do {
    dpp' \<leftarrow> fcc_proc I f fcs Pb' R' dpp
      <+? (\<lambda>s. i +@+
        '': error when applying the flat context closure processor on the DP problem'' +#+
        shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
    check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
      <+? (\<lambda>s. i +@+ '': error below the flat context closure processor'' +#+ shows_nl
        +@+ indent s)
  })"
| "check_dp_proof assms i I dpp (SplitProc Prem Rrem prf1 prf2) = debug (i []) ''SplitProc'' (case get_fcc_option prf1 of 
          Some (f,fcs,Pb',Rb',prf1') \<Rightarrow> debug (i []) ''SplitProcFcc'' (do {
            (dpp1,dpp2) \<leftarrow> fcc_split_proc I f fcs Pb' Rb' Prem Rrem dpp;
            check_dp_proof assms (i +@+ shows_string ''.1.1'') I dpp1 prf1'
              <+? (\<lambda>s. i +@+ '': error below the split and fcc processor'' +#+ shows_nl
                +@+ indent s);
            check_dp_proof assms (i +@+ shows_string ''.2'') I dpp2 prf2
              <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
                 +@+ indent s)
          })
       | None \<Rightarrow> do {
    let (dpp1,dpp2) = split_proc I dpp Prem Rrem;
    check_dp_proof assms (i +@+ shows_string ''.1'') I dpp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s);
    check_dp_proof assms (i +@+ shows_string ''.2'') I dpp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s)
  })" 
| "check_dp_proof assms i I dpp (SemlabProc sli lP lQ lR prf) = 
    debug (i []) ''SemlabProc'' (do {
      dpp' \<leftarrow> semlab_fin_proc I sli lP lQ lR dpp
        <+? (\<lambda>s. i +@+ '': error when applying the semlab processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the semlab processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (SwitchInnermostProc joins prf) = 
    debug (i []) ''SwitchInnermostProc'' (do {
      dpp' \<leftarrow> switch_innermost_proc I joins dpp
        <+? (\<lambda>s. i +@+ '': error when applying the processor to switch to innermost on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the switch to innermost processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_proof assms i I dpp (AssumeFinite dpp') =
    debug (i []) ''Finiteness Assumption'' (
      if assms
        then check_dpp_subsumes I dpp' dpp
          <+? (\<lambda>s. i +@+ '': error in finiteness assumption'' +#+ shows_nl +@+ s +@+ shows_nl)
        else error (i +@+ '': the proof contains a local finiteness assumption'' +#+ shows_nl)
    )"
| "check_dp_proof assms i I dpp (RewritingProc U st st' lr p prf) = debug (i []) ''RewritingProc'' (do {
         dpp' \<leftarrow> rewriting_proc I U st st' lr p dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the rewriting processor to rewrite the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pair '' +#+ shows_nl +@+ shows_rule st' +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the rewriting processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (NarrowingProc st p sts prf) = debug (i []) ''NarrowingProc'' (do {
         dpp' \<leftarrow> narrowing_proc I st p sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the processor processor to narrow the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the narrowing processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (InstantiationProc st sts prf) = debug (i []) ''InstantiationProc'' (do {
         dpp' \<leftarrow> instantiation_proc I st sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (ForwardInstantiationProc st sts u_opt prf) = debug (i []) ''ForwardInstantiationProc'' (do {
         dpp' \<leftarrow> forward_instantiation_proc I st sts u_opt dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the forward_instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the forward instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_proof assms i I dpp (UnlabProc P R prf) = debug (i []) ''UnlabProc'' (error (shows ''unlabeling processor not supported''))"
  by pat_completeness auto

termination
proof (rule, unfold update_error_return)
  fix a i I dpp 
    and edpts::"(('f::{show,linorder},'l::{show,linorder}, string) dp_proof option
      \<times> ('f, 'l, string ) trsLL) list" 
    and pdpps pdpp i' prof dpp' j
  assume proc: "dep_graph_proc I dpp edpts = Inr pdpps"
    and mem: "pdpp \<in> set pdpps" and pdpp: "(prof, dpp') = pdpp"
  from proc mem[unfolded pdpp[symmetric]]
    have "Some prof \<in> set (map fst edpts)" unfolding dep_graph_proc_def by force
  then obtain P where mem: "(Some prof, P) \<in> set edpts" by auto
  let ?m = "prod_size (option_size size)
              (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))
                  (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))))"
  have "size prof < ?m (Some prof, P)" by auto
  also have "... \<le> list_size ?m edpts" using mem by (induct edpts, auto)
  finally have "size prof < size (DepGraphProc edpts)" by simp
  thus "((a, i +@+ ''.'' +#+ shows (Suc j), I, dpp', prof), a, i, I, dpp, DepGraphProc edpts)
    \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]" by auto
next
 fix assms i I dpp Prem Rrem prf1 prf2 a x y xa ya xb yb xc yc yd xd ye
 assume ass1: "get_fcc_option prf1 = Some a" and ass2: "(x, y) = a" "(xa, ya) = y"
         "(xb, yb) = ya" "(xc, yc) = yb"
         "fcc_split_proc I x xa xb xc Prem Rrem dpp = Inr yd" "(xd, ye) = yd"
 show "((assms, i +@+ shows_string ''.1.1'', I, xd, yc), assms, i, I,
            dpp, SplitProc Prem Rrem prf1 prf2)
           \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]" 
 proof -
   from ass1 obtain f fcs Pb' Rb' prf1a 
     where prf1: "prf1 = FccProc f fcs Pb' Rb' prf1a" by (cases prf1, force+)
   from ass1 ass2 
   show ?thesis unfolding prf1 by auto
 qed
qed auto

lemma isOK_check_dp_proof_AssumeFinite_False[simp]:
  "isOK (check_dp_proof False i I dpp (AssumeFinite dpp')) = False" by simp


fun dp_assms :: "bool \<Rightarrow> ('f, 'l, 'v) dp_proof \<Rightarrow> ('f, 'l, 'v) dppLL list" where
  "dp_assms a PisEmpty = []"
| "dp_assms a (SubtermCriterionProc _ _ _ p) = dp_assms a p"
| "dp_assms a (RedPairProc _ _ p) = dp_assms a p"
| "dp_assms a (RedPairUrProc _ _ _ p) = dp_assms a p"
| "dp_assms a (UsableRulesProc _ p) = dp_assms a p"
| "dp_assms a (DepGraphProc ps) =
    concat (concat (map (map (dp_assms a) \<circ> option_to_list \<circ> fst) ps))"
| "dp_assms a (MonoRedPairProc _ _ _ p) = dp_assms a p"
| "dp_assms a (MonoRedPairUrProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SizeChangeSubterm _) = []"
| "dp_assms a (SizeChangeRedPair _ _ _) = []"
| "dp_assms a (UncurryProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (FccProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SplitProc _ _ p q) = dp_assms a p @ dp_assms a q"
| "dp_assms a (SemlabProc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (SwitchInnermostProc _ p) = dp_assms a p"
| "dp_assms a (RewritingProc _ _ _ _ _ p) = dp_assms a p"
| "dp_assms a (NarrowingProc _ _ _ p) = dp_assms a p"
| "dp_assms a (InstantiationProc _ _ p) = dp_assms a p"
| "dp_assms a (ForwardInstantiationProc _ _ _ p) = dp_assms a p"
| "dp_assms a (AssumeFinite d) = (if a then [d] else [])"
| "dp_assms a (UnlabProc _ _ p) = dp_assms a p"
| "dp_assms a (QReductionProc _ p) = dp_assms a p"

lemma dp_assms_False[simp]: "dp_assms False prf = []"
  by (induct False "prf" rule: dp_assms.induct) simp_all


lemma check_dp_proof_with_dp_assms:
  assumes I: "dpp_spec I"
    and fin: "\<forall>p\<in>set (dp_assms a prf). finite_dppL p"
    and ok: "isOK (check_dp_proof a i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
using ok I fin
proof (induct a i I dpp "prf" rule: check_dp_proof.induct)
  case (1 a i I dpp)
  note IH = this
  interpret dpp_spec I by fact
  from IH(1) show ?case by (auto simp:
    QDP_Framework.finite_dpp_def dpp_spec_sound QDP_Framework.ichain.simps)
next
  case (2 a i I dpp p rseq rP prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "subterm_criterion_proc I (create_proj p) (create_rseq_map rseq) ?Pr dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF subterm_criterion_proc,
      OF proc IH(1)[OF HOL.refl HOL.refl _ rec IH(3) fin]])
      (unfold update_error_return, rule proc)
next
  case (3 a i I dpp rpc rP prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with IH(2)
    obtain dpp' where proc: "generic_ur_af_redtriple_proc I (get_redtriple rp) None ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin, unfolded Inr update_error_return sum.simps, OF proc]
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_redtriple_proc[OF IH(3) get_redtriple] _ fin], 
        rule proc)
  next
    case (Inl rrp)
    with IH(2)
    obtain dpp' where proc: "generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) None ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin, unfolded Inl update_error_return sum.simps, OF proc] 
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis
      by (rule sound_proc_impl[OF generic_ur_af_root_redtriple_proc[OF IH(3) get_root_redtriple] _ fin], rule proc)
  qed
next
  case (4 a i I dpp ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "usable_rules_proc I ur dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF usable_rules_proc, OF _ proc 4(1)[OF _ rec _ fin]], insert proc 4(3), auto)
next
  case (5 a i I dpp q prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "q_reduction_proc I q dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case      
    by (rule sound_proc_impl[OF q_reduction_proc[OF IH(3)], OF proc IH(1), 
      unfolded update_error_return, OF proc rec IH(3) fin]) 
next
  case (6 a i I dpp rp rP rR prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "mono_redpair_proc I (get_redtriple rp) ?Pr ?Rr dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case      
    by (rule sound_proc_impl[OF mono_redpair_proc[OF get_redtriple],
      OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
      unfold update_error_return, rule proc)
next
  case (7 a i I dpp edpts)
  note IH = this
  interpret dpp_spec I by fact
  from IH(2) obtain pdpps where proc: "dep_graph_proc I dpp edpts = return pdpps" 
    and rec: "\<And>j. j < length pdpps \<Longrightarrow>
      isOK ((\<lambda>(prof, dpp') j.
        check_dp_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) (pdpps ! j) j)"
    by auto
  show ?case
  proof (rule dep_graph_proc[OF IH(3) proc])
    fix p dpp'
    assume mem: "(p, dpp') \<in> set pdpps"
    from this[unfolded set_conv_nth]
      obtain j where j: "j < length pdpps" and pair: "pdpps ! j = (p, dpp')" by auto
    from rec[OF j] have rec: "isOK (check_dp_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' p)"
      unfolding pair by simp
    from proc[unfolded dep_graph_proc_def, simplified]
      have pdpps: "pdpps =
      map (\<lambda>aP. (the (fst aP), dpp_ops.intersect_pairs I dpp (snd aP)))
          [aP\<leftarrow>edpts . \<exists>y. fst aP = Some y]" by simp
    from mem and IH(4) and proc
      have fin: "\<forall>a\<in>set (dp_assms a p). finite_dppL a" by (simp add: dep_graph_proc_def) force
    show "finite_dpp (dpp_ops.dpp I dpp')" 
      by (rule IH(1)[unfolded update_error_return, OF proc mem HOL.refl rec IH(3) fin])
  qed
next
  case (8 a i I dpp rpc rP ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with IH(2) obtain dpp'
      where proc: "generic_ur_af_redtriple_proc I (get_redtriple rp) (Some ur) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_redtriple_proc[OF IH(3) get_redtriple],
        OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
        unfold Inr update_error_return sum.simps, rule proc)
  next
    case (Inl rrp)
    with IH(2) obtain dpp'
      where proc: "generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) (Some ur) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_proof a ?i I dpp' prof)" 
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_root_redtriple_proc[OF IH(3) get_root_redtriple],
        OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
        unfold Inl update_error_return sum.simps, rule proc)
  qed
next
  case (9 a i I dpp rp rP rR ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "mono_ur_redpair_proc I (get_redtriple rp) ?Pr ?Rr ur dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF mono_ur_redpair_proc[OF get_redtriple],
      OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
      unfold update_error_return, rule proc)
next
  case (10 a i I dpp mode u_info P' R' prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain
    dpp' where proc: "uncurry_proc_both I mode u_info P' R'  dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case      
    by (rule sound_proc_impl[OF uncurry_proc_both[OF IH(3)], OF proc IH(1)],
      unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (11 a i I dpp graphs)
  note IH = this
  interpret dpp_spec I by fact
  from IH(1) have proc: "isOK (sct_subterm_proc I graphs dpp)" by auto
  show ?case
    by (rule sct_subterm_proc_sound[OF proc])
next
  case (12 a i I dpp rp u_opt graphs)
  note IH = this
  interpret dpp_spec I by fact
  from IH(1) have proc: "isOK (sct_ur_af_proc I (get_redtriple rp) graphs u_opt dpp)" by auto
  show ?case
    by (rule sct_ur_af_proc[OF IH(2) get_redtriple, OF proc])
next
  case (13 a i I dpp f fcs Pb' R' prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "fcc_proc I f fcs Pb' R' dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)" by auto
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  from IH(1)[OF _ rec IH(3) fin, simplified, OF proc] have "finite_dpp (dpp_ops.dpp I dpp')"
    by (simp add: dpp_spec_sound)
  from sound_proc_impl[OF fcc_proc, OF proc this] show ?case .
next
  case (14 a i I dpp P' R' prof1 prof2)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "\<lambda>s. i +@+ shows_string s"
  let ?i1 = "?i ''.1''"
  let ?i11 = "?i ''.1.1''"
  let ?i2 = "?i ''.2''"
  note res = IH(5)[unfolded check_dp_proof.simps debug.simps]
  note ass = IH(7)[unfolded dp_assms.simps]
  show ?case
  proof (cases "get_fcc_option prof1")
    case None
    obtain dpp1 dpp2 where s: "split_proc I dpp P' R' = (dpp1, dpp2)" by force
    note ss = s[symmetric]
    note res = res[unfolded None s option.simps Let_def split]
    from res have ok1: "isOK (check_dp_proof a ?i1 I dpp1 prof1)"
    and ok2: "isOK (check_dp_proof a ?i2 I dpp2 prof2)" by auto
    from ass have fin1: "\<forall>p\<in>set (dp_assms a prof1). finite_dppL p"
      and fin2: "\<forall>p\<in>set (dp_assms a prof2). finite_dppL p" by simp_all
    from IH(1)[OF None ss refl ok1 IH(6) fin1]
    have fin1': "finite_dpp (dpp_ops.dpp I dpp1)" .
    from IH(2)[OF None ss refl _ ok2 IH(6) fin2] ok1
    have fin2': "finite_dpp (dpp_ops.dpp I dpp2)" by auto
    show ?thesis
      by (rule split_proc[OF s fin1' fin2'])
  next
    case (Some fcco)
    from Some obtain f fcs Pb' Rb' prf1a where prof1: "prof1 = FccProc f fcs Pb' Rb' prf1a" by (cases prof1, force+)
    hence fcco: "get_fcc_option prof1 = Some (f,fcs,Pb',Rb',prf1a)" by simp
    note res = res[unfolded option.simps fcco Let_def split]
    let ?call = "fcc_split_proc I f fcs Pb' Rb' P' R' dpp"
    from res obtain dpp1 dpp2 where proc: "?call = Inr (dpp1,dpp2)" 
      by (cases ?call, auto)
    note res = res[unfolded proc, simplified]
    from res have ok1: "isOK (check_dp_proof a ?i11 I dpp1 prf1a)"
      and ok2: "isOK (check_dp_proof a ?i2 I dpp2 prof2)" by auto
    from ass have fin1: "\<forall>p\<in>set (dp_assms a prf1a). finite_dppL p"
      and fin2: "\<forall>p\<in>set (dp_assms a prof2). finite_dppL p" 
      unfolding prof1 by simp_all
    from IH(3)[OF fcco refl refl refl refl proc refl ok1 IH(6) fin1]
    have fin1': "finite_dpp (dpp_ops.dpp I dpp1)" .
    from IH(4)[OF fcco refl refl refl refl proc refl _ ok2 IH(6) fin2] ok1
    have fin2': "finite_dpp (dpp_ops.dpp I dpp2)" by auto
    show ?thesis
      by (rule fcc_split_proc[OF proc fin1' fin2'])
  qed
next
  case (15 a i I dpp sli lP lQ lR prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "semlab_fin_proc I sli lP lQ lR dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case
    by (rule sound_proc_impl[OF semlab_fin_proc[OF IH(3)], OF proc IH(1)],
      unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (16 a i I dpp joins prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "switch_innermost_proc I joins dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF switch_innermost_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (17 a i I dpp dpp')
  note IH = this
  interpret dpp_spec I by fact
  have "a = True \<or> a = False" by simp
  thus ?case
  proof
    assume False: "a = False" thus ?thesis using IH(1) by simp
  next
    assume True: "a = True"
    show ?thesis
    proof (cases dpp')
      case fields
      show ?thesis using IH[unfolded True,simplified] and check_dpp_subsumes_sound by best
    qed
  qed
next
  case (18 a i I dpp U st st' lr p prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "rewriting_proc I U st st' lr p dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF rewriting_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (19 a i I dpp st p sts prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "narrowing_proc I st p sts dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF narrowing_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (20 a i I dpp st sts prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "instantiation_proc I st sts dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF instantiation_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (21 a i I dpp st sts u_opt prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "forward_instantiation_proc I st sts u_opt dpp = return dpp'"
    and rec: "isOK (check_dp_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dppL p" by simp
  show ?case    
    by (rule sound_proc_impl[OF forward_instantiation_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (22 a i I dpp P R prof)
  thus ?case by simp
qed

(*the old soundness lemma is an instance of the new one*)
lemma check_dp_proof:
  assumes "dpp_spec I"
    and "isOK (check_dp_proof False i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  using check_dp_proof_with_dp_assms[OF assms(1) _ assms(2)] by simp

(* provide a method to eliminate unlab *)
fun update where "update p_to_p (p,info) = (p_to_p p,info)"

definition "no_decr \<equiv> filter (\<lambda> (l,r). \<not> (l \<noteq> r \<and> map_term unlab l = map_term unlab r))"

definition "unlab_of_trs r \<equiv> unlab_rules (no_decr r)"

fun unlab_to_split_proc :: "('f, 'l, 'v) dp_proof \<Rightarrow> ('f, 'l, 'v) dp_proof \<times> (('f,'l,'v)trsLL \<times> ('f,'l,'v)trsLL \<times> ('f,'l,'v) dp_proof) list" where
  "unlab_to_split_proc PisEmpty = (PisEmpty,[])"
| "unlab_to_split_proc (DepGraphProc ps) = (DepGraphProc
   (map (\<lambda> (po,P). (Option.map (fst o unlab_to_split_proc) po,P)) ps), [])" 
| "unlab_to_split_proc (SubtermCriterionProc a1 a2 a3 p) = update (SubtermCriterionProc a1 a2 a3) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (RedPairProc a1 a2 p) = update (RedPairProc a1 a2) (unlab_to_split_proc p)"
| "unlab_to_split_proc (RedPairUrProc a1 a2 a3 p) = update (RedPairUrProc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (UsableRulesProc a1 p) = update (UsableRulesProc a1) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (QReductionProc a1 p) = update (QReductionProc a1) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (MonoRedPairProc a1 a2 a3 p) = update (MonoRedPairProc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (MonoRedPairUrProc a1 a2 a3 a4 p) = update (MonoRedPairUrProc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (SizeChangeSubterm a1) = (SizeChangeSubterm a1, [])"
| "unlab_to_split_proc (SizeChangeRedPair a1 a2 a3) = (SizeChangeRedPair a1 a2 a3, [])"
| "unlab_to_split_proc (UncurryProc a1 a2 a3 a4 p) = update (UncurryProc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (FccProc a1 a2 a3 a4 p) = update (FccProc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (SwitchInnermostProc a1 p) = update (SwitchInnermostProc a1) (unlab_to_split_proc p)"
| "unlab_to_split_proc (RewritingProc a1 a2 a3 a4 a5 p) = update (RewritingProc a1 a2 a3 a4 a5) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (NarrowingProc a1 a2 a3 p) = update (NarrowingProc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (InstantiationProc a1 a2 p) = update (InstantiationProc a1 a2) (unlab_to_split_proc p)"
| "unlab_to_split_proc (ForwardInstantiationProc a1 a2 a3 p) = update (ForwardInstantiationProc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (AssumeFinite d) = (AssumeFinite d, [])"
| "unlab_to_split_proc (UnlabProc P R p) = (PisEmpty,[(P,R,p)])"
| "unlab_to_split_proc (SplitProc a1 a2 p q) = (SplitProc a1 a2 (fst (unlab_to_split_proc p)) (fst (unlab_to_split_proc q)), [])" 
| "unlab_to_split_proc (SemlabProc a1 lP a2 lR p) = (
     case unlab_to_split_proc p of
        (p,list) \<Rightarrow> let sl = SemlabProc a1 lP a2 lR p
        in case list of 
              Nil \<Rightarrow> (sl,list)
            | Cons (P,R,prof) list' \<Rightarrow> 
          (let ulP = unlab_of_trs lP; ulR = unlab_of_trs lR; PR = list_diff ulP P; RR = list_diff ulR R
            in (SplitProc PR RR sl prof,list'))
     )"  

lemma unlab_to_split_proc_assms: "set (dp_assms a (fst (unlab_to_split_proc prf))) \<subseteq> set (dp_assms a prf)"
proof -
  let ?a = "dp_assms a"
  let ?ap = "\<lambda> p. set (?a p)"
  let ?aps = "\<lambda> pl. ?ap (fst pl) \<union> \<Union> set (map (\<lambda> (_,_,p). ?ap p) (snd pl))"
  have "?aps (unlab_to_split_proc prf) \<subseteq> ?ap prf"
  proof (induct "prf" rule: unlab_to_split_proc.induct)
    case 1 thus ?case by simp
  next
    case (2 ps)
    note IH = this
    let ?m = "(\<lambda>(po,y). (Option.map (fst o unlab_to_split_proc) po,y))"
    let ?ps = "map ?m ps"
    let ?f = "(\<lambda>(po,y). \<Union> (?ap ` Option.set po))"
    have "?aps (unlab_to_split_proc (DepGraphProc ps)) = ?ap (DepGraphProc ?ps)" by simp
    also have "... = \<Union> (set (map ?f ?ps))" 
      by force 
    also have "... \<subseteq> \<Union> (set (map ?f ps))"
    proof(rule subsetI)
      fix r
      assume "r \<in> \<Union> (set (map ?f ?ps))"
      then obtain x where r: "r \<in> ?f x" and "x \<in> set ?ps" by force+
      then obtain y where x: "x = ?m y" and mem: "y \<in> set ps"  by auto
      obtain y1 y2 where y: "y = (y1,y2)" by force
      note x = x[unfolded y split]
      have "?f x \<subseteq> ?f y" unfolding y x split using IH[OF mem y[symmetric]]
        by (cases y1, auto)
      with mem have "?f x \<subseteq> \<Union> (set (map ?f ps))" by auto
      with r show "r \<in> \<Union> (set (map ?f ps))" by blast
    qed
    also have "... = ?ap (DepGraphProc ps)" by force
    finally show ?case .
  next
    case (3 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (4 a1 a2 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (5 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (6 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (7 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (8 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (9 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (10 a1)
    thus ?case by simp
  next
    case (11 a1)
    thus ?case by simp
  next
    case (12 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (13 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (14 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (15 a1 a2 a3 a4 a5 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (16 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (17 a1 a2 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (18 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (19 a1)
    thus ?case by simp
  next
    case (20 P R p)
    thus ?case by simp
  next
    case (21 a1 a2 p q)
    thus ?case by (cases "unlab_to_split_proc p", cases "unlab_to_split_proc q", auto)
  next
    case (22 a1 lP a2 lR p)
    note IH = this
    obtain p' list where up: "unlab_to_split_proc p = (p',list)" by force
    show ?case
    proof (cases list)
      case Nil
      show ?thesis using IH
        unfolding unlab_to_split_proc.simps up fst_conv snd_conv split Let_def Nil
        by auto
    next
      case (Cons prP list')
      obtain P R p'' where prP: "prP = (P,R,p'')" by (cases prP, auto)
      show ?thesis using IH
        unfolding unlab_to_split_proc.simps up fst_conv snd_conv split Let_def Cons prP list.simps by auto
    qed
  qed
  thus ?thesis by auto
qed

hide_const Check_DPP.update

definition check_dp_proof_with_unlab where 
  "check_dp_proof_with_unlab a i I dpp prf \<equiv> check_dp_proof a i I dpp (fst (unlab_to_split_proc prf))"
 
lemma check_dp_proof_with_unlab_dp_assms:
  assumes I: "dpp_spec I"
    and fin: "\<forall>p\<in>set (dp_assms a prf). finite_dppL p"
    and ok: "isOK (check_dp_proof_with_unlab a i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  by (rule check_dp_proof_with_dp_assms[OF I _ ok[unfolded check_dp_proof_with_unlab_def]], insert fin unlab_to_split_proc_assms, auto)

lemma check_dp_proof_with_unlab:
  assumes "dpp_spec I"
    and "isOK (check_dp_proof_with_unlab False i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  using check_dp_proof_with_unlab_dp_assms[OF assms(1) _ assms(2)] by simp

end