theory Linear_Poly_Complexity
imports Complexity
begin

locale linear_poly_carrier =  
  fixes wf_elem :: "'a::show \<Rightarrow> bool"
  and zero :: "'a" ("0")
  and plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "+" 65)
  and times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "*" 70)
  and geq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 21)
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix ">" 21)
  and plus_single_mono :: "bool"
  assumes wf_zero: "wf_elem 0"
  and wf_plus: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> wf_elem (x+y)"
  and plus_comm: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> x+y = y+x"
  and plus_ass: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x+(y+z) = (x+y)+z"
  and zero_left: "wf_elem x \<Longrightarrow> 0+x = x"
  and wf_times: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> wf_elem (x*y)"
  and times_ass: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y*z) = (x*y)*z"
  and left_mult_zero: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> 0*x = 0"
  and right_mult_zero: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> x*0 = 0"
  and distrib_left: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y+z) = (x*y)+(x*z)"
  and distrib_right: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> (x+y)*z = (x*z)+(y*z)"
  and compat: "\<lbrakk>s \<ge> (t :: 'a); t > u; wf_elem s; wf_elem t; wf_elem u\<rbrakk> \<Longrightarrow> s > u"
  and compat2: "\<lbrakk>s > (t :: 'a); t \<ge> u; wf_elem s; wf_elem t; wf_elem u\<rbrakk> \<Longrightarrow> s > u"
  and plus_left_mono: "\<lbrakk>x \<ge> y; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + z \<ge> y + z"
  and times_left_mono: "\<lbrakk>z \<ge> 0; x \<ge> y; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * z \<ge> y * z"
  and times_right_mono: "\<lbrakk>x \<ge> 0; y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * y \<ge> x * z"
  and geq_refl: "wf_elem x \<Longrightarrow> x \<ge> x"
  and geq_trans[trans]: "\<lbrakk>x \<ge> y; y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x \<ge> z"
  and gt_trans[trans]: "\<lbrakk>x > y; y > z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x > z"
  and plus_gt_left_mono: "\<lbrakk>x > y; wf_elem x; wf_elem y; wf_elem z; plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + z"
  and plus_gt_both_mono: "\<lbrakk>x > y; z > u; wf_elem x; wf_elem y; wf_elem z; wf_elem u; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u"
  and times_gt_left_mono: "\<lbrakk>x > y; wf_elem x; wf_elem y; wf_elem z; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x * z > y * z"
  and times_gt_right_mono: "\<lbrakk>x > y; wf_elem x; wf_elem y; wf_elem z; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> z * x  > z * y"
  and zero_leastI: "\<lbrakk>wf_elem x; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x > 0" 
  and zero_leastII: "\<lbrakk>0 > x; wf_elem x; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x = 0" 

locale linear_poly_order_carrier = linear_poly_carrier + 
  fixes one :: "'a :: show" ("1")
   and  default :: "'a"
   and  max0 :: "'a \<Rightarrow> 'a"
   and  arc_pos :: "'a \<Rightarrow> bool"
  assumes SN: "SN {(x,y) . wf_elem x \<and> wf_elem y \<and> geq y zero \<and> arc_pos y \<and> gt x y}"
    and max0_id_pos: "\<lbrakk>geq x  zero; wf_elem x\<rbrakk> \<Longrightarrow> max0 x = x"
    and max0_pos: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) zero"
    and max0_mono: "\<lbrakk>x \<ge> y; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> geq (max0 x) (max0 y)"
    and max0_x: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> geq (max0 x) x"
    and wf_max0: "wf_elem x \<Longrightarrow> wf_elem (max0 x)"
    and wf_one: "wf_elem one"
    and wf_default: "wf_elem default"
    and left_one: "wf_elem x \<Longrightarrow> one * x = x"
    and right_one: "wf_elem x \<Longrightarrow> x * one = x"
    and one_geq_zero: "one \<ge> zero"
    and default_geq_zero: "default \<ge> zero"
    and arc_pos_plus: "\<lbrakk>arc_pos x; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> arc_pos (x + y)"
    and arc_pos_mult: "\<lbrakk>arc_pos x; arc_pos y; wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> arc_pos (x * y)"
    and arc_pos_max0: "wf_elem x \<Longrightarrow> arc_pos (max0 x) = arc_pos x"
    and arc_pos_default: "arc_pos default"
    and max0_npsm_id: "\<not> plus_single_mono \<Longrightarrow> max0 = id"
    and arc_pos_one: "arc_pos one"
    and arc_pos_zero: "\<not> plus_single_mono \<Longrightarrow> \<not> arc_pos zero"
    and not_all_ge: "\<And> c d. \<not> plus_single_mono \<Longrightarrow> wf_elem c \<Longrightarrow> wf_elem d \<Longrightarrow> arc_pos d \<Longrightarrow> \<exists> e. wf_elem e \<and> geq e zero \<and> arc_pos e \<and> \<not> geq c (times d e)"


locale mono_linear_poly_order_carrier = linear_poly_order_carrier +
  fixes check_mono :: "'a :: show \<Rightarrow> bool"
     and max :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  assumes plus_single_mono: "plus_single_mono"
  and check_mono: "\<lbrakk>check_mono x; gt y z; geq x zero; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> gt (times x y) (times x z)" 
  and default_gt_zero: "gt default zero"
  and arc_pos_True: "arc_pos x"
  and wf_max: "wf_elem x \<Longrightarrow> wf_elem y \<Longrightarrow> wf_elem (max x y)"
  and max_left: "wf_elem x \<Longrightarrow> wf_elem y \<Longrightarrow> geq (max x y) x"
  and max_right: "wf_elem x \<Longrightarrow> wf_elem y \<Longrightarrow> geq (max x y) y"

locale complexity_linear_poly_order_carrier = mono_linear_poly_order_carrier + 
  fixes bound :: "'a :: show \<Rightarrow> nat"
  assumes bound: "deriv_bound {(x,y) . wf_elem x \<and> wf_elem y \<and> geq y zero \<and> arc_pos y \<and> gt x y} x (bound x)"
  and bound_plus: "wf_elem x \<Longrightarrow> wf_elem y \<Longrightarrow> bound (plus x y) \<le> bound x + bound y"
  and bound_0[simp]: "bound zero = 0"
  and bound_mono: "x \<ge> y \<Longrightarrow> wf_elem x \<Longrightarrow> wf_elem y \<Longrightarrow> bound x \<ge> bound y"

end