theory Matrix_Complexity
imports Matrix Fact Complexity SN_Order_Carrier
begin

fun mat_powI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> nat \<Rightarrow> 'a mat"
  where "mat_powI ze on pl ti n m 0 = mat1I ze on n"
      | "mat_powI ze on pl ti n m (Suc i) = mat_multI ze pl ti n (mat_powI ze on pl ti n m i) m"

lemma mat_pow: assumes "mat n n m"
  shows "mat n n (mat_powI ze on pl ti n m i)"
proof (induct i)
  case 0
  show ?case unfolding mat_powI.simps by (rule mat1)
next
  case (Suc i)
  show ?case unfolding mat_powI.simps
    by (rule mat_mult[OF Suc assms])
qed

fun upper_triangularI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "upper_triangularI ze on g [] = True"
      | "upper_triangularI ze on g ((a # as) # m) = (g on a \<and> (\<forall> b \<in> set as. b = ze) \<and> upper_triangularI ze on g (map tl m))"
      | "upper_triangularI ze on g ([] # m) = False"

lemma upper_triangular: assumes "mat n n m"
  shows "upper_triangularI ze on g m = (\<forall> i < n. g on (m ! i ! i) \<and> (\<forall> j < n. j > i \<longrightarrow> m ! i ! j = ze))"
  using assms
proof (induct n arbitrary: m)
  case 0
  thus ?case unfolding mat_def by auto
next
  case (Suc n)
  note mat = Suc(2)
  note mat = mat[unfolded mat_def]
  from mat obtain r rs where m: "m = r # rs" by (cases m, auto)
  from mat m obtain a as where r: "r = a # as" by (cases r, auto simp: vec_def)
  let ?r = "map hd rs" 
  let ?m = "map tl rs"
  from mat[unfolded m r] have l: "length ?m = n" "length ?r = n" "length rs = n" and v: "list_all (vec (Suc n)) rs" "vec (Suc n) (a # as)" by auto
  from v have "list_all (vec n) ?m" by (induct rs, auto simp: vec_def)
  with l have mat: "mat n n ?m" unfolding mat_def by simp
  from v have las: "length as = n" unfolding vec_def by simp
  note IH = Suc(1)[OF mat]
  let ?l = "\<lambda> i. g on (map tl rs ! i ! i) \<and>
            (\<forall>j<n. i < j \<longrightarrow> map tl rs ! i ! j = ze)"
  let ?r = "\<lambda> i. g on (rs ! i ! Suc i) \<and>
            (\<forall>j<n. i < j \<longrightarrow> rs ! i ! Suc j = ze)"
  {
    fix i
    assume i: "i < n"
    with l have irs:  "i < length rs" by auto
    from v have "\<forall> r \<in> set rs. vec (Suc n) r" unfolding list_all_iff by auto
    from this[unfolded all_set_conv_all_nth l] i have v: "vec (Suc n) (rs ! i)" by auto
    hence "length (rs ! i) = Suc n" unfolding vec_def by auto
    then obtain a as where rsi: "rs ! i = a # as" and las: "length as = n" by (cases "rs ! i", auto)
    have "?l i = ?r i" 
      unfolding nth_map[OF irs] 
      unfolding rsi by auto      
  }
  hence main: "(\<forall>i<n. ?l i) = (\<forall> i<n. ?r i)" by auto
  show ?case unfolding m r upper_triangularI.simps
    unfolding IH 
    unfolding all_Suc 
    unfolding all_set_conv_all_nth l las
    by (auto simp: main)
qed    

context ordered_semiring_1
begin

lemma scalar_prod: "scalar_prod v1 v2 = listsum (map (\<lambda>(x,y). x * y) (zip v1 v2))" 
proof -
  obtain z where z: "zip v1 v2 = z" by auto
  show ?thesis unfolding scalar_prodI_def listsum_def z 
    by (induct z, auto)
qed

abbreviation mat_pow where "mat_pow \<equiv> mat_powI (0 :: 'a) 1 (op +) (op *)"
abbreviation upper_triangular where "upper_triangular \<equiv> upper_triangularI (0 :: 'a) 1 ge
"

lemma pow_mono: assumes ab: "a \<succeq> b" and b: "b \<succeq> 0"
  shows "a ^ n \<succeq> b ^ n \<and> b ^ n \<succeq> 0"
proof (induct n)
  case 0
  show ?case by (auto simp: ge_refl one_ge_zero)
next
  case (Suc n)
  hence abn: "a ^ n \<succeq> b ^ n" and bn: "b ^ n \<succeq> 0" by auto
  have bsn: "b ^ Suc n \<succeq> 0" unfolding power_Suc
    using times_left_mono[OF bn b] by auto
  have "a ^ Suc n = a * a ^ n" unfolding power_Suc by simp
  also have "... \<succeq> b * a ^ n"
    by (rule times_left_mono[OF ge_trans[OF abn bn] ab])
  also have "... \<succeq> b * b ^ n"
    by (rule times_right_mono[OF b abn])
  finally show ?case using bsn unfolding power_Suc by simp
qed

lemma pow_ge_zero[intro]: assumes a: "a \<succeq> (0 :: 'a)"
  shows "a ^ n \<succeq> 0"
proof (induct n)
  case 0
  from one_ge_zero show ?case by simp
next
  case (Suc n)
  show ?case using times_left_mono[OF Suc a] by simp
qed
end

lemma listsum_0: "\<lbrakk>\<And> x. x \<in> set xs \<Longrightarrow> x = 0\<rbrakk> \<Longrightarrow> listsum xs = 0"
  by (induct xs, auto)

lemma of_nat_ge_zero[intro,simp]: "of_nat n \<succeq> (0 :: ('a :: ordered_semiring_1))"
proof (induct n)
  case 0
  show ?case by (simp add: ge_refl)
next
  case (Suc n)
  from plus_right_mono[OF Suc, of 1] have "of_nat (Suc n) \<succeq> (1 :: 'a)" by simp
  also have "... \<succeq> 0" using one_ge_zero .
  finally show ?case .
qed

lemma mult_ge_zero[intro]: "(a :: 'a :: ordered_semiring_1) \<succeq> 0 \<Longrightarrow> b \<succeq> 0 \<Longrightarrow> a * b \<succeq> 0"
  using times_left_mono[of b a 0] by auto

lemma upper_triangular_mat_pow_main: assumes mat: "mat d d (m :: ('a :: {comm_semiring_1, ordered_semiring_1}) mat)"
  and tri: "upper_triangular m"
  and a: "\<And> i j. i < d \<Longrightarrow> j < d \<Longrightarrow> ge (m ! i ! j) 0 \<and> ge a (m ! i ! j)"
  and a1: "ge a 1"
  shows "\<forall> i < d. \<forall> j < d. (i > j \<longrightarrow> mat_pow d m n ! j ! i = 0) \<and> ge (of_nat (fact (j - i)) * (of_nat n * a)^(j - i)) (mat_pow d m n ! j ! i)"
proof -
  let ?n = "of_nat :: nat \<Rightarrow> 'a"
  let ?m = "mat_pow d m"
  let ?p = "\<lambda> n i j. i > j \<longrightarrow> ?m n ! j ! i = 0"
  let ?q = "\<lambda> n i j. ge (?n (fact (j - i)) * (?n n * a)^(j - i)) (?m n ! j ! i)"
  show "\<forall> i < d. \<forall> j < d. ?p n i j \<and> ?q n i j"
  proof (induct n)
    case 0
    show ?case 
    proof (intro allI impI)
      fix i j
      assume i: "i < d" and j: "j < d"
      have m0: "?m 0 = mat1 d" by simp
      note id = m0 mat1_index[OF j i]
      show "?p 0 i j \<and> ?q 0 i j" unfolding id by (cases "j - i", auto intro: ge_refl one_ge_zero)
    qed
  next
    case (Suc n)
    show ?case
    proof (intro allI impI)
      fix i j
      assume i: "i < d" and j: "j < d"
      have mS: "?m (Suc n) = mat_mult d (?m n) m" by simp
      note id = mS mat_mult_to_scalar[OF mat_pow[OF mat] mat i j]
      let ?rowi = "row (?m n) i"
      let ?colj = "col m j"
      let ?prod = "scalar_prod ?rowi ?colj"
      from row[OF mat_pow[OF mat] i] have row: "vec d ?rowi" .
      from col[OF mat j] have col: "vec d ?colj" .
      have id: "?m (Suc n) ! j ! i = ?prod" unfolding id ..
      let ?z = "zip ?rowi ?colj" 
      from row col have len: "length ?rowi = length ?colj" and lenz: "length ?z = d" unfolding vec_def by auto
      let ?map = "map (\<lambda> k. (?m n ! k ! i, m ! j ! k)) [0 ..< d]"
      note tri = tri[unfolded upper_triangular[OF mat], rule_format]
      have "?z = map (\<lambda> k. (?rowi ! k, ?colj ! k)) [0 ..< d]"
        unfolding zip_nth_conv[OF len] using col[unfolded vec_def] by simp
      also have "... = ?map"
        unfolding map_eq_conv
      proof
        fix k
        assume "k \<in> set [0..<d]"
        hence k: "k < d" by auto
        show "(?rowi ! k, ?colj ! k) = (?m n ! k ! i, m ! j ! k)"
          unfolding row_col[OF mat_pow[OF mat] i k]
          unfolding col_def by simp
      qed
      finally have zip: "?z = ?map" .
      {
        fix k
        assume k: "k < d"
        hence kd: "k < length [0..<d]" and nth: "[0..<d] ! k = k" by auto
        hence "?z ! k = (?m n ! k ! i, m ! j ! k)"
          unfolding zip nth_map[OF kd] nth by simp
      } note zk = this        
      {
        assume ji: "j < i"
        obtain z where z: "?z = z" by auto        
        have "\<forall> ab \<in> set ?z. fst ab = 0 \<or> snd ab = 0"  
          unfolding all_set_conv_all_nth lenz
        proof (intro allI impI)
          fix k
          assume k: "k < d"
          show "fst (?z ! k) = 0 \<or> snd (?z ! k) = 0" 
            unfolding zk[OF k] fst_conv snd_conv
          proof (cases "k < i")
            case True
            from Suc[rule_format, THEN conjunct1, rule_format, OF i k True]
            show "?m n ! k ! i = 0 \<or> m ! j ! k = 0" by simp
          next
            case False
            with ji k i j have "j < k" by arith
            from tri[OF j, THEN conjunct2, rule_format, OF k this] 
            show "?m n ! k ! i = 0 \<or> m ! j ! k = 0" by simp
          qed
        qed
        hence "\<forall> (a,b) \<in> set z. a = 0 \<or> b = 0" unfolding z by force
        hence "?prod = 0" unfolding scalar_prod z 
          by (induct z, auto)
      } note lower_part = this
      hence p: "?p (Suc n) i j" unfolding id by simp
      from a[OF i i] have a0: "a \<succeq> 0" using ge_trans[of a "m ! i ! i" 0] by auto
      have na: "\<And> n. ?n n * a \<succeq> (0 :: 'a)"
        by (rule mult_ge_zero[OF of_nat_ge_zero a0])
      have "?q (Suc n) i j"
      proof (cases "j < i") 
        case True
        show ?thesis unfolding id lower_part[OF True] 
          by (rule mult_ge_zero[OF of_nat_ge_zero pow_ge_zero[OF na]])
      next
        case False
        hence ji: "j \<ge> i" by simp
        let ?fact = "\<lambda> k. ?n (fact (k - i)) * (?n n * a) ^ (k - i)" 
        let ?fac = "\<lambda> k. ?n (fact k) * (?n n * a) ^ k"
        let ?ind = "\<lambda> k. (if k < i then 0 else ?fact k)"
        let ?mind = "map ?ind [0..< d]"
        have mind: "vec d ?mind" unfolding vec_def by auto
        let ?fsn = "?n (fact (j - i)) * (?n (Suc n) * a) ^ (j - i)"
        let ?cola = "\<lambda> k. if k < j then a else if k = j then 1 else 0"
        let ?mcola = "map ?cola [0 ..< d]"
        let ?both = "\<lambda> k. (?ind k, ?cola k)"
        let ?prod = "\<lambda> k. (?ind k * ?cola k)"
        let ?mboth = "map ?both [0 ..< d]" 
        let ?mprod = "map ?prod [0 ..< d]" 
        have cola: "vec d ?mcola" unfolding vec_def by simp
        let ?z = "zip ?mind ?mcola"
        have len: "length ?mind = length ?mcola" by simp
        have z: "?z = ?mboth"
          unfolding zip_nth_conv[OF len] by auto
        let ?f = "(\<lambda> (x,y). op + (x * y))"
        let ?id = "map ?prod [i ..< d]" 
        let ?ij = "map ?prod [i ..< Suc j]" 
        let ?jd = "map ?prod [Suc j ..< d]"
        let ?zi = "map ?prod [0 ..< i]" 
        obtain i_d where i_d: "i_d = ?id" by auto
        obtain diff where diff: "j - i = diff" by auto
        have "scalar_prod ?mind ?mcola = listsum ?mprod" unfolding scalar_prod unfolding z unfolding map_map o_def by simp
        also have "... = listsum ?zi + listsum i_d"
          using upt_add_eq_append[of 0 i "d - i"] i i_d by auto
        also have "listsum ?zi = 0"
        proof (rule listsum_0)
          fix x
          assume "x \<in> set ?zi"
          then obtain k where k: "k < i" and xy: "x = ?prod k" by auto
          from k xy show "x = 0" by auto
        qed
        also have "0 + listsum i_d = listsum ?id" unfolding i_d  by simp
        also have "?id = ?ij @ ?jd"
          using upt_add_eq_append[of i "Suc j" "d - Suc j"] ji j by auto
        also have "listsum (?ij @ ?jd) = listsum ?ij + listsum ?jd" by simp
        also have "listsum ?jd = 0"
        proof (rule listsum_0)
          fix x
          assume "x \<in> set ?jd"
          then obtain k where k: "Suc j \<le> k" "k < d" and xy: "x = ?prod k" by auto
          from k xy show "x = 0" by auto
        qed
        also have "listsum ?ij + 0 = listsum ?ij" by simp
        also have "?ij = map (\<lambda> k. (?fact k * ?cola k)) [i ..< Suc j]"
          by (rule map_cong, auto)
        also have "... = map (\<lambda> k. (?fact k * ?cola k)) [i ..< j] @ [?fact j]"
          using upt_add_eq_append[of i j "Suc 0", OF ji]  by auto
        also have "... = map (\<lambda> k. (?fact k * a)) [i ..< j] @ [?fact j]" (is "_ = ?zwi @ _") by auto
        also have "?zwi = map (\<lambda> k. (?fac k * a)) [0 ..< j - i]" (is "_ = ?map")
          by (rule nth_map_conv, auto)     
        also have "listsum (?map @ [?fact j]) = listsum ?map + ?fact j" by simp
        finally have sprod: "scalar_prod ?mind ?mcola = listsum ?map + ?fact j" .
        have "?fsn \<succeq> scalar_prod ?mind ?mcola" 
          unfolding sprod
          unfolding diff
          unfolding of_nat_Suc
        proof (induct diff)
          case 0
          show ?case by (simp add: ge_refl)
        next
          case (Suc d)
          note IH = this
          show ?case
          proof (cases d)
            case 0
            show ?thesis unfolding 0 by (auto simp: ge_refl field_simps)
          next
            case (Suc dd)                        
            have ana0: "(1 + ?n n) * a \<succeq> 0"
              unfolding of_nat_Suc[symmetric] by (rule na)
            have anana: "(1 + ?n n) * a \<succeq> ?n n * a"
              using plus_right_mono[OF a0, of "a * ?n n"] by (auto simp: field_simps)
            have ananapow: "((1 + ?n n) * a) ^ d \<succeq> (?n n * a) ^ d \<and> (?n n * a)^d \<succeq> 0"
              by (rule pow_mono[OF anana na])
            have "?n (fact (Suc d)) * (a + ?n n * a) ^ Suc d = 
              ?n (fact (Suc d)) * ((a + ?n n * a) * (a + ?n n * a) ^ d)" 
              unfolding power_Suc by simp
            also have "... = ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              ?n (fact d) * (a * (a + ?n n * a) ^ d)" 
              by (simp add: field_simps)
            also have "... \<succeq> ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              ?n (fact d) * (1 * (a + ?n n * a) ^ d)" 
              by (rule plus_right_mono[OF times_right_mono[OF of_nat_ge_zero times_left_mono[OF pow_ge_zero a1]]], insert na[of "Suc n"], auto simp: field_simps)
            also have "?n (fact d) * (1 * (a + ?n n * a) ^ d) = ?n (fact d) * (((1 + ?n n) * a) ^ d)" by (simp add: field_simps)
            also have "?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              ?n (fact d) * (((1 + ?n n) * a) ^ d) \<succeq> 
              ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              ((\<Sum>k\<leftarrow>[0..<d]. ?n (fact k) * (?n n * a) ^ k * a) +
  ?n (fact d) * (?n n * a) ^ d)"
              by (rule plus_right_mono[OF IH])
            also have "... = ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              (\<Sum>k\<leftarrow>[0..<d]. ?n (fact k) * (?n n * a) ^ k * a) +
                ?n (fact d) * (?n n * a) ^ d" by (simp add: field_simps)
            also have "... \<succeq> ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d) + 
              (\<Sum>k\<leftarrow>[0..<d]. ?n (fact k) * (?n n * a) ^ k * a) +
                0"
              by (rule plus_right_mono[OF mult_ge_zero[OF of_nat_ge_zero]], insert ananapow, auto)
            also have "... = ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              (\<Sum>k\<leftarrow>[0..<d]. ?n (fact k) * (?n n * a) ^ k * a) +
              ?n (d * fact d) * (a * (a + ?n n * a) ^ d)" by (simp add: ac_simps)
            also have "... \<succeq> ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d) + 
              (\<Sum>k\<leftarrow>[0..<d]. ?n (fact k) * (?n n * a) ^ k * a) + 
              ?n (fact d) * (?n n * a) ^ d * a"
            proof (rule plus_right_mono)
              have "?n (d * fact d) * (a * (a + ?n n * a) ^ d)
                = ?n (fact d) * (a * (a + ?n n * a) ^ d) + 
                ?n (dd * fact d) * (a * (a + ?n n * a) ^ d)" unfolding Suc  by (simp add: field_simps)
              also have "... \<succeq> ?n (fact d) * (a * (a + ?n n * a) ^ d) + 0"
                by (rule plus_right_mono[OF mult_ge_zero[OF of_nat_ge_zero mult_ge_zero[OF a0]]], insert ana0, auto simp: field_simps)
              also have "... = (?n (fact d) * (a + ?n n * a) ^ d) * a"
                by (auto simp: field_simps)
              also have "... \<succeq> (?n (fact d) * (?n n * a) ^ d) * a"
                by (rule times_left_mono[OF a0 times_right_mono[OF of_nat_ge_zero]], insert ananapow, auto simp: field_simps)
              finally show "?n (d * fact d) * (a * (a + ?n n * a) ^ d) \<succeq> (?n (fact d) * (?n n * a) ^ d) * a" .
            qed
            also have "... = (\<Sum>k\<leftarrow>[0..< Suc d]. ?n (fact k) * (?n n * a) ^ k * a) 
               + ?n (fact (Suc d)) * ((?n n * a) * (a + ?n n * a) ^ d)"
              by (simp add: field_simps)
            also have "... \<succeq> (\<Sum>k\<leftarrow>[0..<Suc d]. ?n (fact k) * (?n n * a) ^ k * a) +
              ?n (fact (Suc d)) * (?n n * a) ^ Suc d" unfolding power_Suc
              by (rule plus_right_mono[OF times_right_mono[OF of_nat_ge_zero times_right_mono[OF na]]], insert ananapow, auto simp: field_simps)
            finally
            show "?n (fact (Suc d)) * ((1 + ?n n) * a) ^ Suc d \<succeq>
              (\<Sum>k\<leftarrow>[0..<Suc d]. ?n (fact k) * (?n n * a) ^ k * a) +
              ?n (fact (Suc d)) * (?n n * a) ^ Suc d" by (simp add: field_simps)
          qed
        qed      
        also have "... \<succeq> scalar_prod ?mind ?colj"
        proof (rule scalar_prod_mono_right[OF mind cola col])
          show "vec_ge ?mind (vec0 d)"
            unfolding vec_ge_index[OF mind vec0] using a0
            by (auto simp: vec0I_def intro: mult_ge_zero ge_refl)
        next
          show "vec_ge ?mcola ?colj" 
            unfolding vec_ge_index[OF cola col] 
            unfolding col_def 
          proof (intro allI impI)
            fix i
            assume i: "i < d"
            hence id: "?mcola ! i = ?cola i" by auto
            have "?cola i \<succeq> m ! j ! i"
            proof (cases "i < j")
              case True
              with a[OF j i] show ?thesis by auto
            next
              case False
              hence ij: "i \<ge> j" by simp
              show ?thesis
              proof (cases "i = j")
                case True
                with tri[OF j] show ?thesis by auto
              next
                case False
                with ij have ji: "j < i" by auto
                with tri[OF j] i show ?thesis by (auto simp: ge_refl)
              qed
            qed
            thus "?mcola ! i \<succeq> m ! j ! i" unfolding id .
          qed
        qed
        also have "... \<succeq> scalar_prod ?rowi ?colj"
        proof (rule scalar_prod_mono_left[OF mind row col])
          show "vec_ge ?colj (vec0 d)" 
            unfolding vec_ge_index[OF col vec0]
            unfolding col_def vec0I_def using a[OF j] by auto
        next
          show "vec_ge ?mind ?rowi"
            unfolding vec_ge_index[OF mind row]
          proof (intro allI impI)
            fix j
            assume j: "j < d"
            have idr: "?rowi ! j = ?m n ! j ! i" unfolding row_col[OF mat_pow[OF mat] i j] col_def ..
            have idl: "?mind ! j = ?ind j" using j by auto
            note IH = Suc[rule_format, OF i j]
            have "?ind j \<succeq> ?m n ! j ! i" 
              by (cases "j < i", insert IH, auto simp: ge_refl)
            thus "?mind ! j \<succeq> ?rowi ! j" unfolding idl idr .
          qed
        qed
        finally
        show "?fsn \<succeq> ?m (Suc n) ! j ! i" unfolding id .
      qed
      with p 
      show "?p (Suc n) i j \<and> ?q (Suc n) i j" ..
    qed
  qed
qed

lemma of_nat_mono[intro]: assumes "n \<ge> m" shows "(of_nat n :: 'a :: ordered_semiring_1)  \<succeq> of_nat m"
proof -
  let ?n = "of_nat :: nat \<Rightarrow> 'a"
  from assms
  show ?thesis
  proof (induct m arbitrary: n)
    case 0
    show ?case by auto
  next
    case (Suc m nn)
    then obtain n where nn: "nn = Suc n" by (cases nn, auto)
    note Suc = Suc[unfolded nn]
    hence rec: "?n n \<succeq> ?n m" by simp
    show ?case unfolding nn of_nat_Suc
      by (rule plus_right_mono[OF rec])
  qed
qed

lemma pow_mono_one: assumes a: "a \<succeq> (1 :: 'a :: ordered_semiring_1)"
  shows "a ^ n \<succeq> 1"
proof (induct n)
  case (Suc n)
  show ?case unfolding power_Suc
    using ge_trans[OF times_right_mono[OF ge_trans[OF a one_ge_zero] Suc], of 1]
    a
    by (auto simp: field_simps)
qed (auto simp: ge_refl)

lemma pow_mono_exp: assumes a: "a \<succeq> (1 :: 'a :: ordered_semiring_1)"
  shows "n \<ge> m \<Longrightarrow> a ^ n \<succeq> a ^ m"
proof (induct m arbitrary: n)
  case 0
  show ?case using pow_mono_one[OF a] by auto
next
  case (Suc m nn)
  then obtain n where nn: "nn = Suc n" by (cases nn, auto)
  note Suc = Suc[unfolded nn]
  hence rec: "a ^ n \<succeq> a ^ m" by auto
  show ?case unfolding nn power_Suc
    by (rule times_right_mono[OF ge_trans[OF a one_ge_zero] rec])
qed

lemma mult_ge_one[intro]: assumes a: "(a :: 'a :: ordered_semiring_1) \<succeq> 1"
  and b: "b \<succeq> 1"
  shows "a * b \<succeq> 1"
proof -
  from ge_trans[OF b one_ge_zero] have b0: "b \<succeq> 0" .
  from times_left_mono[OF b0 a] have "a * b \<succeq> b" by simp
  from ge_trans[OF this b] show ?thesis .
qed

fun vec_fold :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a vec \<Rightarrow> 'b \<Rightarrow> 'b"
  where [code_unfold]: "vec_fold f = foldr f"

fun mat_fold :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a mat \<Rightarrow> 'b \<Rightarrow> 'b"
  where [code_unfold]: "mat_fold f = foldr (vec_fold f)"

lemma foldr_foldr_concat: "foldr (foldr f) m a = foldr f (concat m) a"
proof (induct m arbitrary: a)
  case Nil show ?case by simp
next
  case (Cons v m a)
  show ?case unfolding concat.simps foldr.simps Cons 
    unfolding foldr_append by simp
qed

definition linear_norm :: "('a :: monoid_add)mat \<Rightarrow> 'a"
  where "linear_norm m \<equiv> listsum (concat m)"

lemma concat_mat: "mat nr nc m \<Longrightarrow>
  concat m = [ m ! i ! j. i \<leftarrow> [0 ..< nc], j \<leftarrow> [0 ..< nr] ]"
proof (induct m arbitrary: nc)
  case Nil
  thus ?case unfolding mat_def by auto
next
  case (Cons v m snc)
  from Cons(2) obtain nc where snc: "snc = Suc nc" and mat: "mat nr nc m" and v: "vec nr v"
    unfolding mat_def by (cases snc, auto)
  from v have nr: "nr = length v" unfolding vec_def by auto
  have v: "map (\<lambda> i. v ! i) [0 ..< nr] = v" unfolding nr map_nth by simp
  note IH = Cons(1)[OF mat]
  show ?case 
    unfolding snc 
    unfolding map_nth_Suc
    unfolding nth.simps nat.simps concat.simps
    unfolding IH v ..
qed

lemma listsum_ge_mono: fixes as :: "('a :: ordered_semiring_0) list"
  assumes "length as = length bs"
  and "\<And> i. i < length bs \<Longrightarrow> as ! i \<succeq> bs ! i"
  shows "listsum as \<succeq> listsum bs"
  using assms 
proof (induct as arbitrary: bs)
  case (Nil bs)
  from Nil(1) show ?case by (simp add: ge_refl)
next
  case (Cons a as bbs)
  from Cons(2) obtain b bs where bbs: "bbs = b # bs" and len: "length as = length bs" by (cases bbs, auto)
  note ge = Cons(3)[unfolded bbs]
  {
    fix i
    assume "i < length bs"
    hence "Suc i < length (b # bs)" by simp
    from ge[OF this] have "as ! i \<succeq> bs ! i" by simp
  }
  from Cons(1)[OF len this] have IH: "listsum as \<succeq> listsum bs" .
  from ge[of 0] have ab: "a \<succeq> b" by simp
  from ge_trans[OF plus_left_mono[OF ab] plus_right_mono[OF IH]]
  show ?case unfolding bbs  by simp
qed

lemma vec_ge_listsum: fixes v1 :: "('a :: ordered_semiring_0) vec"
  assumes v1: "vec nr v1" and v2: "vec nr v2" and ge: "vec_ge v1 v2"
  shows "listsum v1 \<succeq> listsum v2" 
proof -
  from v1 v2 have len: "length v1 = length v2" "length v2 = nr" unfolding vec_def by auto
  show ?thesis
    by (rule listsum_ge_mono[OF len(1), unfolded len(2)], insert ge[unfolded vec_ge_index[OF v1 v2]], auto)
qed

lemma linear_norm_ge: fixes m1 :: "('a :: ordered_semiring_0) mat"
  assumes m1: "mat nr nc m1" and m2: "mat nr nc m2"
  and ge: "mat_ge m1 m2" 
  shows "linear_norm m1 \<succeq> linear_norm m2"
  using assms 
proof (induct m1 arbitrary: m2 nc)
  case Nil
  thus ?case unfolding mat_def by (cases m2, auto simp: ge_refl)
next
  case (Cons v1 m1 m2v snc)
  from Cons(2) Cons(3) obtain nc v2 m2 where m2v: "m2v = v2 # m2" and m1: "mat nr nc m1" and m2: "mat nr nc m2" and v1: "vec nr v1" and v2: "vec nr v2" and snc: "snc = Suc nc" unfolding mat_def by (cases m2v, auto)
  note Cons = Cons[unfolded this]
  from Cons(4) have v12': "vec_ge v1 v2" and m12ge: "mat_ge m1 m2"
    unfolding mat_ge_def mat_comp_all_def vec_ge_def by auto
  from vec_ge_listsum[OF v1 v2 v12'] have v12: "listsum v1 \<succeq> listsum v2" .
  from Cons(1)[OF  m1 m2 m12ge] have m12: "linear_norm m1 \<succeq> linear_norm m2" .
  from ge_trans[OF plus_left_mono[OF v12] plus_right_mono[OF m12]]
  have vm12: "linear_norm (v1 # m1) \<succeq> linear_norm (v2 # m2)" unfolding linear_norm_def by simp
  thus ?case unfolding m2v .
qed

context one_mono_ordered_semiring_1
begin

lemma vec_gt_listsum: fixes v1 :: "'a vec"
  assumes "vec nr v1" and "vec nr v2" and "vec_ge v1 v2" and "vec_pre_gtI op \<succ> (sub_vec sd v1) (sub_vec sd v2)"
  shows "listsum v1 \<succ> listsum v2" 
proof -
  note d = vec_def
  from assms
  show ?thesis
  proof (induct v1 arbitrary: v2 nr sd)
    case Nil
    thus ?case unfolding d sub_vec_def vec_pre_gtI_def by auto
  next
    case (Cons a1 v1 av2 snr ssd)
    from Cons(2) Cons(3) obtain a2 v2 nr where av2: "av2 = a2 # v2" and snr: "snr = Suc nr" and v2: "vec nr v2" unfolding d
      by (cases snr, (cases av2, auto)+)
    note Cons = Cons[unfolded av2 snr]
    from Cons(2) have v1: "vec nr v1" unfolding d by simp
    from Cons(4) have a12: "a1 \<succeq> a2" and v12: "vec_ge v1 v2"
      unfolding vec_ge_def vec_comp_all_def by auto
    from Cons(5) obtain sd where "(a1 \<succ> a2) \<or> (vec_pre_gtI op \<succ> (sub_vec sd v1) (sub_vec sd v2))" 
      unfolding vec_pre_gtI_def sub_vec_def by (cases ssd, auto)
    thus ?case unfolding av2
    proof 
      assume "a1 \<succ> a2"
      from compat[OF plus_right_mono[OF vec_ge_listsum[OF v1 v2 v12]] plus_gt_left_mono[OF this]] show "listsum (a1 # v1) \<succ> listsum (a2 # v2)" by simp
    next
      assume "vec_pre_gtI op \<succ> (sub_vec sd v1) (sub_vec sd v2)"
      from Cons(1)[OF v1 v2 v12 this] have "listsum v1 \<succ> listsum v2" .
      from compat[OF plus_right_mono[OF a12] plus_gt_left_mono[OF this]]
      show "listsum (a1 # v1) \<succ> listsum (a2 # v2)" by (simp add: ac_simps)
    qed
  qed
qed

lemma linear_norm_gt_main: assumes m1: "mat nr nc m1" and m2: "mat nr nc m2"
  and ge: "mat_ge m1 m2" 
  shows "linear_norm m1 \<succeq> linear_norm m2 \<and> (mat_pre_gt (sub_mat sd sdc m1) (sub_mat sd sdc m2) \<longrightarrow> linear_norm m1 \<succ> linear_norm m2)"
proof -
  note d = vec_def
  from ge m1 m2 show "(linear_norm m1 \<succeq> linear_norm m2) \<and> (mat_pre_gt (sub_mat sd sdc m1) (sub_mat sd sdc m2) \<longrightarrow> linear_norm m1 \<succ> linear_norm m2)"
  proof (induct m1 arbitrary: m2 nc sdc)
    case Nil
    thus ?case unfolding mat_pre_gtI_def sub_mat_def mat_def by (cases m2, auto simp: ge_refl)
  next
    case (Cons v1 m1 m2v snc ssdc)
    from Cons(3) Cons(4) obtain nc v2 m2 where m2v: "m2v = v2 # m2" and m1: "mat nr nc m1" and m2: "mat nr nc m2" and v1: "vec nr v1" and v2: "vec nr v2" unfolding mat_def by (cases m2v, auto)
    note Cons = Cons[unfolded this]
    from Cons(2) have v12': "vec_ge v1 v2" and m12ge: "mat_ge m1 m2"
      unfolding mat_ge_def mat_comp_all_def vec_ge_def by auto
    note IH = Cons(1)[OF m12ge m1 m2]
    from vec_ge_listsum[OF v1 v2 v12'] have v12: "listsum v1 \<succeq> listsum v2" .
    from IH have m12: "linear_norm m1 \<succeq> linear_norm m2" by simp
    from ge_trans[OF plus_left_mono[OF v12] plus_right_mono[OF m12]]
    have vm12: "linear_norm (v1 # m1) \<succeq> linear_norm (v2 # m2)" unfolding linear_norm_def by simp
    show ?case unfolding m2v
    proof (rule conjI[OF vm12], rule impI)
      assume gt: "mat_pre_gt (sub_mat sd ssdc (v1 # m1)) (sub_mat sd ssdc (v2 # m2))"
      from gt[unfolded mat_pre_gtI_def sub_mat_def] obtain sdc where ssd: "ssdc = Suc sdc" by (cases ssdc, auto)
      from gt ssd have "mat_pre_gt (sub_mat sd sdc m1) (sub_mat sd sdc m2) \<or> vec_pre_gtI op \<succ> (sub_vec sd v1) (sub_vec sd v2)" unfolding mat_pre_gtI_def sub_mat_def by auto
      thus "linear_norm (v1 # m1) \<succ> linear_norm (v2 # m2)" 
      proof
        assume "mat_pre_gt (sub_mat sd sdc m1) (sub_mat sd sdc m2)"
        with IH[of sdc] have "linear_norm m1 \<succ> linear_norm m2" by auto
        from compat[OF plus_right_mono[OF v12] plus_gt_left_mono[OF this]]
        show ?thesis unfolding linear_norm_def by (simp add: ac_simps)
      next
        assume "vec_pre_gtI op \<succ> (sub_vec sd v1) (sub_vec sd v2)"
        from compat[OF plus_right_mono[OF m12] plus_gt_left_mono[OF vec_gt_listsum[OF v1 v2 v12' this]]]
        show ?thesis unfolding linear_norm_def by (simp add: ac_simps)
      qed
    qed
  qed
qed

lemma linear_norm_gt: assumes m1: "mat nr nc m1" and m2: "mat nr nc m2"
  and gt: "mat_gt sd m1 m2" 
  shows "linear_norm m1 \<succ> linear_norm m2"
proof -
  from gt[unfolded mat_gtI_def] have ge: "mat_ge m1 m2"
    and gt: "mat_pre_gt (sub_mat sd sd m1) (sub_mat sd sd m2)" by auto
  from linear_norm_gt_main[OF m1 m2 ge] gt show ?thesis by auto
qed
end

lemma linear_norm_0: "linear_norm (mat0 nr nc) = (0 :: 'a :: comm_monoid_add)"
  unfolding linear_norm_def mat_fold.simps mat0I_def
proof (induct nc)
  case 0 show ?case by simp
next
  case (Suc n)
  show ?case 
    unfolding replicate.simps foldr.simps concat.simps listsum_append Suc
    unfolding vec0I_def by (induct nr, auto)
qed

lemma linear_norm_ge_0: fixes m :: "('a :: ordered_semiring_0) mat"
  assumes m: "mat nr nc m" 
  and ge: "mat_ge m (mat0 nr nc)" 
  shows "linear_norm m \<succeq> 0"
  using linear_norm_ge[OF m mat0 ge]
  unfolding linear_norm_0 .

lemma linear_norm_plus: "mat nr nc m1 \<Longrightarrow> mat nr nc (m2 :: ('a :: comm_monoid_add)mat) \<Longrightarrow> linear_norm (mat_plus m1 m2) = linear_norm m1 + linear_norm m2" 
  unfolding linear_norm_def mat_fold.simps 
proof (induct nc arbitrary: m1 m2)
  case 0 thus ?case unfolding mat_def mat_plusI_def by simp
next
  case (Suc nc vm1 vm2)
  from Suc(2) obtain v1 m1 where vm1: "vm1 = v1 # m1" and m1: "mat nr nc m1" and v1: "vec nr v1" unfolding mat_def by (cases vm1, auto)
  from Suc(3) obtain v2 m2 where vm2: "vm2 = v2 # m2" and m2: "mat nr nc m2" and v2: "vec nr v2" unfolding mat_def by (cases vm2, auto)
  note IH = Suc(1)[OF m1 m2]
  have "listsum (concat (mat_plus vm1 vm2)) = listsum (vec_plus v1 v2) + listsum (concat (mat_plus m1 m2))" unfolding vm1 vm2 mat_plusI_def by simp
  also have "listsum (vec_plus v1 v2) = listsum v1 + listsum v2"
    using v1 v2
  proof (induct nr arbitrary: v1 v2)
    case 0
    thus ?case unfolding vec_def vec_plusI_def by auto
  next
    case (Suc nr vm1 vm2)
    from Suc(2) obtain v1 m1 where vm1: "vm1 = v1 # m1" and m1: "vec nr m1" unfolding vec_def by (cases vm1, auto)
    from Suc(3) obtain v2 m2 where vm2: "vm2 = v2 # m2" and m2: "vec nr m2" unfolding vec_def by (cases vm2, auto)
    from Suc(1)[OF m1 m2] show ?case unfolding vm1 vm2
      unfolding vec_plusI_def by (auto simp: ac_simps)
  qed
  finally show ?case unfolding IH unfolding vm1 vm2 
    by (simp add: ac_simps)
qed    

lemma linear_norm_index: fixes m :: "('a :: ordered_semiring_1) mat"   
  assumes bc: "bc \<succeq> 0"
  shows "mat nr nc m \<Longrightarrow> \<lbrakk>\<And> i j. i < nc \<Longrightarrow> j < nr \<Longrightarrow> bc \<succeq> m ! i ! j\<rbrakk> \<Longrightarrow> of_nat nr * of_nat nc * bc \<succeq> linear_norm m"
proof (induct nc arbitrary: m)
  case 0 thus ?case using bc unfolding mat_def linear_norm_def 
    by (simp add: ge_refl)
next
  case (Suc nc vm)
  from Suc(2) obtain v m where vm: "vm = v # m" 
    and m: "mat nr nc m" and v: "vec nr v" unfolding mat_def  by (cases vm, auto)
  note Suc = Suc[unfolded vm]
  from Suc(3)[of 0] have bcv: "\<And> j. j < nr \<Longrightarrow> bc \<succeq> v ! j" by simp
  {
    fix i 
    assume "i < nc" 
    with Suc(3)[of "Suc i"] have "\<And> j. j < nr \<Longrightarrow> bc \<succeq> m ! i ! j" by auto
  } note bcm = this
  from Suc(1)[OF m bcm] have IH: "of_nat nr * of_nat nc * bc \<succeq> linear_norm m" by auto
  from v bcv have v: "of_nat nr * bc \<succeq> listsum v"
  proof (induct nr arbitrary: v)
    case 0
    thus ?case unfolding vec_def using bc by (auto simp: ge_refl)
  next
    case (Suc nr av)
    from Suc(2) obtain a v where av: "av = a # v" 
      and v: "vec nr v" unfolding vec_def  by (cases av, auto)
    note Suc = Suc[unfolded av]
    from Suc(3)[of 0] have a: "bc \<succeq> a" by auto
    {
      fix i
      assume "i < nr"
      with Suc(3)[of "Suc i"] have "bc \<succeq> v ! i" by auto 
    } note bcv = this
    from Suc(1)[OF v bcv] have IH: "of_nat nr * bc \<succeq> listsum v" by auto
    have "of_nat (Suc nr) * bc = bc + of_nat nr * bc" by (simp add: field_simps)
    also have "... \<succeq> a + listsum v"
      by (rule ge_trans[OF plus_left_mono[OF a] plus_right_mono[OF IH]])
    finally show ?case unfolding av by simp
  qed
  have "of_nat nr * of_nat (Suc nc) * bc = of_nat nr * bc + of_nat nr * of_nat nc * bc" by (simp add: field_simps)
  also have "... \<succeq> listsum v + linear_norm m"
    by (rule ge_trans[OF plus_left_mono[OF v] plus_right_mono[OF IH]])
  finally show ?case unfolding vm linear_norm_def by auto
qed

(* TODO: move to Matrix *)
lemma col_index: assumes m: "mat nr nc m"
  and i: "i < nc"
  shows "col m i = map (\<lambda> j. m ! i ! j) [0 ..< nr]"
proof -
  from m[unfolded mat_def list_all_iff] i
  have nr: "nr = length (m ! i)" by (auto simp: vec_def)
  show ?thesis unfolding nr col_def 
    by (rule map_nth[symmetric])
qed

(* TODO: move to Matrix *)
lemma row_index: assumes m: "mat nr nc m"
  and i: "i < nr"
  shows "row m i = map (\<lambda> j. m ! j ! i) [0 ..< nc]"
proof -
  note rc = row_col[OF m i]
  from row[OF m i] have id: "length (row m i) = nc" unfolding vec_def by simp
  from map_nth[of "row m i"]
  have "row m i = map (\<lambda> j. row m i ! j) [0 ..< nc]" unfolding id by simp
  also have "... = map (\<lambda> j. m ! j ! i) [0 ..< nc]" using rc[unfolded col_def] by auto
  finally show ?thesis .
qed

lemma listsum_ge_0_nth: fixes xs :: "('a :: ordered_semiring_0)list"
  assumes ge: "\<And> i. i < length xs \<Longrightarrow> xs ! i \<succeq> 0"
  shows "listsum xs \<succeq> 0"
proof -
  let ?l = "replicate  (length xs) (0 :: 'a)"
  have "length xs = length ?l" by simp
  from listsum_ge_mono[OF this] ge have "listsum xs \<succeq> listsum ?l" by simp
  also have "... = 0" using listsum_0[of ?l] by auto
  finally show ?thesis .
qed

lemma listsum_ge_0: fixes xs :: "('a :: ordered_semiring_0)list"
  assumes ge: "\<And> x. x \<in> set xs \<Longrightarrow> x \<succeq> 0"
  shows "listsum xs \<succeq> 0"
  by (rule listsum_ge_0_nth, insert ge[unfolded set_conv_nth], auto)

lemma listsum_double_concat: 
  fixes f :: "'b \<Rightarrow> 'c \<Rightarrow> 'a :: comm_monoid_add" and g as bs
  shows "listsum (concat (map (\<lambda> i. map (\<lambda> j. f i j + g i j) as) bs))
      = listsum (concat (map (\<lambda> i. map (\<lambda> j. f i j) as) bs)) + 
        listsum (concat (map (\<lambda> i. map (\<lambda> j. g i j) as) bs))"
proof (induct bs)
  case Nil thus ?case by simp
next
  case (Cons b bs)
  have id: "(\<Sum>j\<leftarrow>as. f b j + g b j) = listsum (map (f b) as) + listsum (map (g b) as)"
    by (induct as, auto simp: ac_simps)
  show ?case unfolding map.simps concat.simps listsum_append
    unfolding Cons
    unfolding id 
    by (simp add: ac_simps)
qed


lemma linear_norm_submultiplicative: fixes m1 :: "('a :: ordered_semiring_1) mat"
  shows "mat_ge m1 (mat0 nr n) \<Longrightarrow> mat_ge m2 (mat0 n nc) \<Longrightarrow> mat nr n m1 \<Longrightarrow> mat n nc m2 \<Longrightarrow>
  linear_norm m1 * linear_norm m2 \<succeq> linear_norm (mat_mult nr m1 m2)" 
proof (induct n arbitrary: m1 m2)
  case 0
  have m2: "\<forall> x \<in> set m2. x = [] \<Longrightarrow> m2 = replicate (length m2) []"
    by (induct m2, auto)
  with 0(3) 0(4) have m1: "m1 = []" and m2: "m2 = replicate nc []" unfolding mat_def  
    by (auto simp: vec_def list_all_iff)
  have "linear_norm (mat_mult nr m1 m2) = listsum (concat (replicate nc (replicate nr (0\<Colon>'a))))" unfolding m1 m2
    by (simp add: linear_norm_def mat_multI_def matT_vec_multI_def scalar_prodI_def) 
  also have "... = 0"
  proof (induct nc)
    case (Suc nc)
    show ?case by (simp add: Suc, induct nr, auto)
  qed simp
  finally have id: "linear_norm (mat_mult nr m1 m2) = 0" .
  show ?case unfolding id unfolding m1 m2 linear_norm_def by (simp add: ge_refl)
next
  case (Suc n m1 m2)
  let ?mgen = "\<lambda> nr nc m. [ m ! i ! j. i \<leftarrow> [0 ..< nc], j \<leftarrow> [0 ..< nr] ]"
  let ?mn1 = "?mgen nr n"
  let ?msn1 = "?mgen nr (Suc n)"
  let ?mn2 = "?mgen n nc"
  let ?msn2 = "?mgen (Suc n) nc"
  note m1 = Suc(4)
  note m2 = Suc(5)
  note m10 = Suc(2)
  note m20 = Suc(3)
  note IH = Suc(1)
  let ?gen = "\<lambda> (nr :: nat) nc (f :: nat \<Rightarrow> nat \<Rightarrow> 'a). map (\<lambda>i. map (f i) [0..<nr]) [0..<nc]"
  let ?v = "linear_norm"
  let ?ii = "\<lambda> m nr nc. ?gen nr nc (\<lambda> i j. m ! i ! j)"
  obtain ii1 where ii1: "ii1 = ?ii m1 nr n" by auto
  obtain ii2 where ii2: "ii2 = ?ii m2 n nc" by auto
  let ?midx = "\<lambda> m nr nc. ?gen nr nc (\<lambda> i j. m ! i ! j)" 
  obtain nn where nn: "nn = Suc n" by auto
  note nns = nn[symmetric]
  {
    fix nr :: nat and nc :: nat and f :: "nat \<Rightarrow> nat \<Rightarrow> 'a" 
    have "mat nr nc (?gen nr nc f)"
      unfolding mat_def list_all_iff by (auto simp: vec_def)    
  } note gen = this
  {
    fix f and g  :: "nat \<Rightarrow> 'a" and idx
    have "listsum (concat (map (\<lambda> i. f i @ [g i]) idx)) = listsum (concat (map f idx)) + listsum (map g idx)" by (induct idx, auto simp: ac_simps)
  } note listsum_concat_singleton = this
  {
    fix m :: "'a mat" and nr nc
    assume m: "mat nr nc m"
    have "m = ?midx m nr nc" 
      unfolding mat_eq_index[OF m gen] 
      by auto
  } note mat_idx = this
  let ?mni = "[ m1 ! n ! i . i \<leftarrow> [0..<nr]]"
  let ?min = "[ m2 ! i ! n . i \<leftarrow> [0..<nc]]"
  let ?lmni = "listsum ?mni"
  let ?lmin = "listsum ?min"
  {    
    from concat_mat[OF m1]
    have "concat m1 = concat (?midx m1 nr (Suc n))" by simp
    hence "?v m1 = listsum (concat (?midx m1 nr (Suc n)))" unfolding linear_norm_def by simp
    also have "... = listsum (concat (?midx m1 nr n)) + listsum ?mni"
      by simp
    also have "listsum (concat (?midx m1 nr n)) = linear_norm ii1"
      unfolding ii1 linear_norm_def ..
    finally have "linear_norm m1 = linear_norm ii1 + ?lmni" .
  } note vm1 = this
  {    
    from concat_mat[OF m2]
    have "concat m2 = concat (?midx m2 (Suc n) nc)" by simp
    hence "?v m2 = listsum (concat (?midx m2 (Suc n) nc))" unfolding linear_norm_def by simp
    also have "... = listsum (concat (map (\<lambda>i. map (op ! (m2 ! i)) [0..<n] @ [m2 ! i ! n]) [0..<nc]))" by simp
    also have "... = listsum (concat (?midx m2 n nc)) + ?lmin" unfolding listsum_concat_singleton ..
    also have "listsum (concat (?midx m2 n nc)) = linear_norm ii2"
      unfolding ii2 linear_norm_def ..
    finally have "linear_norm m2 = linear_norm ii2 + ?lmin" .
  } note vm2 = this
  have mat_ii1: "mat nr n ii1" unfolding ii1 by (rule gen)
  have mat_ii2: "mat n nc ii2" unfolding ii2 by (rule gen)
  {
    fix i j
    assume i: "i < n" and j: "j < nr"
    have "ii1 ! i ! j = m1 ! i ! j" using i j unfolding ii1 by simp
  } note ii1_idx = this
  {
    fix i j
    assume i: "i < nc" and j: "j < n"
    have "ii2 ! i ! j = m2 ! i ! j" using i j unfolding ii2 by simp
  } note ii2_idx = this
  {
    fix m :: "'a mat" and nr nc
    assume m: "mat nr nc m" and ge: "mat_ge m (mat0 nr nc)"
    from ge[unfolded mat_ge_index[OF m mat0] mat0_index]
    have ge0: "\<And> i j. i < nc \<Longrightarrow> j < nr \<Longrightarrow> m ! i ! j \<succeq> 0" using mat0_index[of _ nc _ nr "0 :: 'a"]
      by auto
  } note mat_ge = this
  have ii10: "mat_ge ii1 (mat0 nr n)"
    unfolding mat_ge_index[OF mat_ii1 mat0] using ii1_idx mat_ge[OF m1 m10] mat0_index[of _ n _ nr "0 :: 'a"]
    by auto
  have ii20: "mat_ge ii2 (mat0 n nc)"
    unfolding mat_ge_index[OF mat_ii2 mat0] using ii2_idx mat_ge[OF m2 m20] mat0_index[of _ nc _ n "0 :: 'a"]
    by auto
  let ?mult = "mat_mult nr m1 m2"
  have m12: "mat nr nc ?mult" by (rule mat_mult[OF m1 m2])
  {
    fix i
    assume i: "i < nc"
    have "col m2 i = col ii2 i @ [m2 ! i ! n]"
      unfolding col_index[OF m2 i]
      unfolding col_index[OF mat_ii2 i]
      using ii2_idx[OF i]
      by auto
  } note col_n = this
  {
    fix i
    assume i: "i < nr"
    have "row m1 i = row ii1 i @ [m1 ! n ! i]"
      unfolding row_index[OF m1 i]
      unfolding row_index[OF mat_ii1 i]
      using ii1_idx[OF _ i]
      by auto
  } note row_n = this
  let ?scalar = "\<lambda> i j. scalar_prod (row m1 j) (col m2 i)"
  let ?scalar' = "\<lambda> i j. scalar_prod (row ii1 j) (col ii2 i) + 
    m1 ! n ! j * m2 ! i ! n" 
  obtain scalar' where scalar': "scalar' = ?scalar'" by auto
  obtain scalar where scalar: "scalar = ?scalar" by auto
  let ?mult' = "?gen nr nc scalar'"
  let ?vmii = "?v (mat_mult nr ii1 ii2)"
  let ?rii = "listsum [m1 ! n ! j * m2 ! i ! n . i \<leftarrow> [0..<nc], j \<leftarrow> [0..<nr]]"
  {
    have mult_mult': "?mult = ?mult'"
      unfolding mat_eq_index[OF m12 gen[of _ _ scalar']]
    proof (intro allI impI)
      fix i j
      assume i: "i < nc" and j: "j < nr"
      have "?mult ! i ! j = scalar i j"
        unfolding mat_mult_to_scalar[OF m1 m2 j i] scalar ..
      also have "... = scalar' i j"
      proof -
        from row[OF mat_ii1 j] col[OF mat_ii2 i]
        have len: "length (row (ii1) j) = length (col (ii2) i)" "length [m1 ! n ! j] = length [m2 ! i ! n]" unfolding vec_def by auto
        show "scalar i j = scalar' i j" 
          unfolding scalar scalar' row_n[OF j] col_n[OF i]
          unfolding scalar_prod
          unfolding zip_append[OF len] by simp
      qed
      finally 
      show "?mult ! i ! j = ?mult' ! i ! j"
         using i j by simp
    qed
    hence "?v ?mult = ?v ?mult'" by simp
    also have "... = listsum (concat (map (\<lambda>i. map (scalar' i) [0..<nr]) [0 ..< nc]))"
      unfolding linear_norm_def ..
    also have "... =
      listsum (concat (map (\<lambda>i. map (\<lambda>j. scalar_prod (row (ii1) j) (col (ii2) i))
                  [0..<nr])
         [0..<nc])) +
    ?rii" (is "_ = ?zwi + _") unfolding scalar' unfolding listsum_double_concat  ..
    also have "?zwi = ?vmii"
      unfolding linear_norm_def
    proof (rule arg_cong[where f = "\<lambda> x. listsum (concat x)"], unfold
      mat_eq_index[OF gen mat_mult[OF mat_ii1 mat_ii2]], intro allI impI)
      fix i j
      assume i: "i < nc" and j: "j < nr"
      let ?sc = "(\<lambda> i j. scalar_prod (row (ii1) j) (col (ii2) i))"
      have "?gen nr nc ?sc ! i ! j = ?sc i j" using i j by simp
      also have "... = mat_mult nr (ii1) (ii2) ! i ! j"
        by (rule mat_mult_to_scalar[symmetric, OF mat_ii1 mat_ii2 j i])
      finally show "?gen nr nc ?sc ! i ! j = mat_mult nr (ii1) (ii2) ! i ! j" .
    qed
    finally have "?v ?mult = ?vmii + ?rii" .
  } note vmult = this
  have "?v m1 * ?v m2 = ?v (ii1) * ?v (ii2) + (?v (ii1) * ?lmin + ?lmni * ?v (ii2) + ?lmni * ?lmin)"
    unfolding vm1 vm2 by (simp add: field_simps)
  also have "... \<succeq> ?vmii + (?v (ii1) * ?lmin + ?lmni * ?v (ii2) + ?lmni * ?lmin)"
    by (rule plus_left_mono[OF IH[OF ii10 ii20 mat_ii1 mat_ii2]])
  also have "... \<succeq> ?vmii + ?rii" 
  proof (rule plus_right_mono)
    from linear_norm_ge_0[OF mat_ii1 ii10] have ii10: "linear_norm ii1 \<succeq> 0" .
    from linear_norm_ge_0[OF mat_ii2 ii20] have ii20: "linear_norm ii2 \<succeq> 0" .
    note m10 = mat_ge[OF m1 m10]
    note m20 = mat_ge[OF m2 m20]
    have lmin0: "?lmin \<succeq> 0"
      by (rule listsum_ge_0_nth, insert m20, auto)
    have lmni0: "?lmni \<succeq> 0"
      by (rule listsum_ge_0_nth, insert m10, auto)
    have p10: "?v (ii1) * ?lmin \<succeq> 0"
      by (rule mult_ge_zero[OF ii10 lmin0])
    have p20: "?lmni * ?v (ii2) \<succeq> 0"
      by (rule mult_ge_zero[OF lmni0 ii20])
    from plus_right_mono[OF ge_trans[OF plus_left_mono[OF p10] plus_right_mono[OF p20]], of "?lmni * ?lmin"]
    have ge: "?v (ii1) * ?lmin + ?lmni * ?v (ii2) + ?lmni * ?lmin \<succeq> ?lmni * ?lmin"
      by (simp add: ac_simps)
    have id: "?lmni * ?lmin = ?rii" 
    proof (induct nc)
      case 0
      show ?case by simp
    next
      case (Suc nc)
      let ?nr = "listsum (map (op ! (m1 ! n)) [0..<nr])"
      let ?nrr = "listsum (concat (map (\<lambda>i. map (\<lambda>j. m1 ! n ! j * m2 ! i ! n) [0..<nr]) [0..<nc]))"
      have "?nr * (\<Sum>i\<leftarrow>[0..<Suc nc]. m2 ! i ! n)
        = ?nr * ((\<Sum>i\<leftarrow>[0..<nc]. m2 ! i ! n) + m2 ! nc ! n)"
        by simp
      also have "... = ?nr * (\<Sum>i\<leftarrow>[0..<nc]. m2 ! i ! n) + ?nr * m2 ! nc ! n" 
        by (simp add: field_simps)
      also have "... = ?nrr + ?nr * m2 ! nc ! n" unfolding Suc ..
      also have "?nr * m2 ! nc ! n = (\<Sum>j\<leftarrow>[0..<nr]. m1 ! n ! j * m2 ! nc ! n)" 
        by (induct nr, auto simp: field_simps)
      also have "?nrr + ... = listsum (concat (map (\<lambda>i. map (\<lambda>j. m1 ! n ! j * m2 ! i ! n) [0..<nr]) [0..<Suc nc]))"
        by simp
      finally
      show ?case .
    qed
    show "?v (ii1) * ?lmin + ?lmni * ?v (ii2) + ?lmni * ?lmin \<succeq> ?rii" 
      using ge unfolding id .
  qed
  also have "... = ?v ?mult" unfolding vmult ..
  finally show ?case .
qed

class large_ordered_semiring_1 = comm_semiring_1 + ordered_semiring_1 +
  assumes ex_large_of_nat: "\<exists> x. of_nat x \<succeq> y"

lemma linear_norm_mult_left_ex: assumes m: "mat n n (m :: ('a :: large_ordered_semiring_1) mat)" 
  and m0: "mat_ge m (mat0 n n)" (is "mat_ge m ?m0")
  shows "\<exists> c. (\<forall> m'. mat n n m' \<longrightarrow> mat_ge m' (mat0 n n) \<longrightarrow> linear_norm m' * (of_nat c) \<succeq> linear_norm (mat_mult n m' m))"
proof -
  let ?c = "linear_norm m"
  from linear_norm_ge_0[OF m m0] have c0: "?c \<succeq> 0" .
  from ex_large_of_nat[of ?c] obtain c where c: "of_nat c \<succeq> ?c" by auto
  show ?thesis
  proof (rule exI[of _ c], intro allI impI)
    fix m' :: "'a mat"
    assume m': "mat n n m'" and m'0: "mat_ge m' (mat0 n n)"
    let ?m' = "linear_norm m'"
    have "?m' * of_nat c \<succeq> ?m' * ?c"
      by (rule times_right_mono[OF linear_norm_ge_0[OF m' m'0] c])
    also have "... \<succeq> linear_norm (mat_mult n m' m)"
      by (rule linear_norm_submultiplicative[OF m'0 m0 m' m])
    finally show "linear_norm m' * of_nat c \<succeq> linear_norm (mat_mult n m' m)" .
  qed
qed

lemma linear_norm_mult_right_ex: assumes m: "mat n n (m :: ('a :: large_ordered_semiring_1) mat)" 
  and m0: "mat_ge m (mat0 n n)" (is "mat_ge m ?m0")
  shows "\<exists> c. (\<forall> m'. mat n n m' \<longrightarrow> mat_ge m' (mat0 n n) \<longrightarrow> linear_norm m' * (of_nat c) \<succeq> linear_norm (mat_mult n m m'))"
proof -
  let ?c = "linear_norm m"
  from linear_norm_ge_0[OF m m0] have c0: "?c \<succeq> 0" .
  from ex_large_of_nat[of ?c] obtain c where c: "of_nat c \<succeq> ?c" by auto
  show ?thesis
  proof (rule exI[of _ c], intro allI impI)
    fix m' :: "'a mat"
    assume m': "mat n n m'" and m'0: "mat_ge m' (mat0 n n)"
    let ?m' = "linear_norm m'"
    have "?m' * of_nat c \<succeq> ?m' * ?c"
      by (rule times_right_mono[OF linear_norm_ge_0[OF m' m'0] c])
    also have "... = ?c * ?m'" by (simp add: ac_simps)
    also have "... \<succeq> linear_norm (mat_mult n m m')"
      by (rule linear_norm_submultiplicative[OF m0 m'0 m m'])
    finally show "linear_norm m' * of_nat c \<succeq> linear_norm (mat_mult n m m')" .
  qed
qed

lemma mat_pow_ge_zero: assumes m: "mat n n (m :: ('a :: ordered_semiring_1) mat)" and m0: "mat_ge m (mat0 n n)"
  shows "mat_ge (mat_pow n m nn) (mat0 n n)"
proof (induct nn)
  case 0
  show ?case by (simp add: mat1_ge_mat0)
next
  case (Suc nn)
  show ?case
    unfolding mat_powI.simps
    by (rule mat_ge_trans[OF mat_mult_left_mono[OF mat_pow[OF m] mat0 m Suc m0] _ mat_mult[OF mat_pow[OF m] m] mat_mult[OF mat0 m] mat0], unfold mat0_mult_left[OF m], rule mat_ge_refl) 
qed    

class max_ordered_ab_semigroup = ordered_ab_semigroup + monoid_add + ord +
  assumes max_ge_x[intro]: "max x y \<succeq> x" 
      and max_ge_y[intro]: "max x y \<succeq> y"

class bin_max_ordered_semiring_1 = large_ordered_semiring_1 + max_ordered_ab_semigroup + max_ordered_semiring_1

locale complexity_one_mono_ordered_semiring_1 = one_mono_ordered_semiring_1 default gt 
  for gt :: "'a :: bin_max_ordered_semiring_1 \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and default :: 'a + 
  fixes bound :: "'a \<Rightarrow> nat"
  assumes bound_mono: "\<And> a b. a \<succeq> b \<Longrightarrow> bound a \<ge> bound b"
   and bound_plus: "\<And> a b. bound (a + b) \<le> bound a + bound b" 
   and bound_plus_of_nat[simp]: "\<And> a n. a \<succeq> 0 \<Longrightarrow> bound (a + of_nat n) = bound a + bound (of_nat n)" 
   and bound_zero[simp]: "bound 0 = 0"
   and bound_one: "bound 1 \<ge> 1"
   and bound: "\<And> a. deriv_bound {(a,b). (b \<succeq> 0) \<and> (a \<succ> b)} a (bound a)"
begin


lemma bound_linear: "\<exists> c. \<forall> n. bound (of_nat n) \<le> c * n"
proof (rule exI[of _ "bound 1"], intro allI)
  fix n
  show "bound (of_nat n) \<le> bound 1 * n"
  proof (induct n)
    case 0
    show ?case by simp
  next
    case (Suc n)
    have "bound (of_nat (Suc n)) = bound (1 + of_nat n)" by simp
    also have "... \<le> bound 1 + bound (of_nat n)"
      by (rule bound_plus)
    also have "... \<le> bound 1 + bound 1 * n"
      using Suc by auto
    finally show ?case by auto
  qed
qed

lemma bound_mult_of_nat: "bound (a * of_nat n) \<le> bound a * bound (of_nat n)"
proof (induct n)
  case 0
  show ?case by simp
next
  case (Suc n)
  have "bound (a * of_nat (Suc n)) = bound (a + a * of_nat n)" by (simp add: field_simps)
  also have "... \<le> bound a + bound (a * of_nat n)"
    by (rule bound_plus)
  also have "... \<le> bound a + bound a * bound (of_nat n)" using Suc by auto
  also have "... = bound a * (1 + bound (of_nat n))" by (simp add: field_simps)
  also have "... \<le> bound a * (bound (1 + of_nat n))"
  proof (rule mult_le_mono2)
    show "1 + bound(of_nat n) \<le> bound (1 + of_nat n)" using bound_one
      unfolding bound_plus_of_nat[OF one_ge_zero] by simp
  qed
  finally show ?case by simp
qed
end

instantiation int :: bin_max_ordered_semiring_1
begin
instance 
proof 
  fix y :: int
  show "\<exists> x. of_nat x \<succeq> y"
    by (rule exI[of _ "nat y"], simp) 
qed auto
end

instantiation rat :: bin_max_ordered_semiring_1
begin
instance 
proof 
  fix y :: rat
  from ex_le_of_nat[of y]
  show "\<exists> x. of_nat x \<succeq> y" by simp
qed auto
end


abbreviation vec_max where "vec_max \<equiv> vec_plusI (max :: 'a :: max_ordered_ab_semigroup \<Rightarrow> 'a \<Rightarrow> 'a)"
abbreviation mat_max where "mat_max \<equiv> mat_plusI (max :: 'a :: max_ordered_ab_semigroup \<Rightarrow> 'a \<Rightarrow> 'a)"

lemmas mat_max_index = mat_plus_index[of _ _ _ _ _ _ max]
lemmas mat_max = mat_plus[of _ _ _ _ max]

definition max_mat where "max_mat m \<equiv> mat_fold max m (0 :: 'a :: max_ordered_ab_semigroup)"

lemma foldr_max: "a \<in> set as \<Longrightarrow> foldr max as b \<succeq> (a :: 'a :: max_ordered_ab_semigroup)"
proof (induct as arbitrary: b)
  case Nil thus ?case by simp
next
  case (Cons c as)
  show ?case
  proof (cases "a = c")
    case True
    show ?thesis unfolding True
      unfolding foldr.simps by auto
  next
    case False
    with Cons have "foldr max as b \<succeq> a" by auto
    from ge_trans[OF _ this] show ?thesis unfolding foldr.simps by auto
  qed
qed

lemma max_mat: assumes m: "mat nr nc m" 
  and i: "i < nc" and j: "j < nr"
  shows "max_mat m \<succeq> m ! i ! j"
proof -
  have id: "max_mat m = foldr max (concat m) 0" 
    unfolding max_mat_def mat_fold.simps vec_fold.simps
    unfolding foldr_foldr_concat ..
  from m[unfolded mat_def list_all_iff] i 
  have v: "\<And> v. v \<in> set m \<Longrightarrow> vec nr v" and mi: "m ! i \<in> set m" by auto
  from v[OF mi] j have mij: "m ! i ! j \<in> set (m ! i)" unfolding vec_def by auto
  from mi mij have mij: "m ! i ! j \<in> set (concat m)" by auto
  show ?thesis unfolding id
    by (rule foldr_max[OF mij])
qed

lemma mat_max_ge: assumes m1: "mat nr nc m1" and m2: "mat nr nc m2"
  shows "mat_ge (mat_max m1 m2) m1" "mat_ge (mat_max m1 m2) m2"
proof -
  have m: "mat nr nc (mat_max m1 m2)"
    by (rule mat_max[OF m1 m2])
  note ind = mat_max_index[OF m1 m2]
  show "mat_ge (mat_max m1 m2) m1"
    unfolding mat_ge_index[OF m m1]
    using ind by auto
  show "mat_ge (mat_max m1 m2) m2"
    unfolding mat_ge_index[OF m m2]
    using ind by auto
qed

definition mat_max_list :: "'a mat list \<Rightarrow> 'a mat \<Rightarrow> ('a :: max_ordered_ab_semigroup) mat"
  where "mat_max_list ms init \<equiv> foldr mat_max ms init"

lemma mat_max_list: assumes init: "mat nr nc init"
  shows "\<lbrakk> \<And> m. m \<in> set ms \<Longrightarrow> mat nr nc m\<rbrakk> \<Longrightarrow> 
  mat nr nc (mat_max_list ms init) \<and> mat_ge (mat_max_list ms init) init \<and> (\<forall> m \<in> set ms. mat_ge (mat_max_list ms init) m)"
  unfolding mat_max_list_def
proof (induct ms)
  case Nil
  thus ?case using init mat_ge_refl by simp
next
  case (Cons m ms)
  let ?m = "mat nr nc"
  from Cons(2) have m: "?m m" and ms: "\<And> m. m \<in> set ms \<Longrightarrow> ?m m" by auto
  let ?mf = "foldr mat_max ms init"
  let ?mm = "mat_max m ?mf"
  note IH = Cons(1)[OF ms]
  from IH have mf: "?m ?mf" and ge_init: "mat_ge ?mf init" and ge_ms: "\<And> m. m \<in> set ms \<Longrightarrow> mat_ge ?mf m" by auto  
  from mat_max[OF m mf] have mm: "?m ?mm" .
  from mat_max_ge[OF m mf] have mm_m: "mat_ge ?mm m" and mm_mf: "mat_ge ?mm ?mf" .
  note mm_mf = mat_ge_trans[OF mm_mf _ mm mf]
  from mm_mf[OF ge_init init] have mm_init: "mat_ge ?mm init" .
  from mm_mf[OF ge_ms ms] have mm_ms: "\<And> m. m \<in> set ms \<Longrightarrow> mat_ge ?mm m" .
  from mm mm_init mm_ms mm_m
  show ?case by simp
qed


lemma upper_triangular_mat_pow_index: assumes mat: "mat d d (m :: ('a :: bin_max_ordered_semiring_1) mat)"
  and tri: "upper_triangular m"
  and ge0: "mat_ge m (mat0 d d)"
  shows "\<exists> c. c \<succeq> 0 \<and> (\<forall> n > 0. \<forall> i < d. \<forall> j < d. ge (c * of_nat n ^ (d - Suc 0)) (mat_pow d m n ! i ! j))"
proof -
  let ?n = "of_nat :: nat \<Rightarrow> 'a"
  let ?d = "d - Suc 0"
  obtain a where a: "a = max 1 (max_mat m)" by auto
  note main = upper_triangular_mat_pow_main[OF mat tri]  
  {
    fix i j
    assume i: "i < d" and j: "j < d"
    from ge_trans[OF _ max_mat[OF mat i j]]
    have "a \<succeq> m ! i ! j" unfolding a by auto
  } note am = this
  have a1: "a \<succeq> 1" unfolding a by auto
  from ge_trans[OF this one_ge_zero] have a0: "a \<succeq> 0" by simp
  from ge0[unfolded mat_ge_index[OF mat mat0]]
  have ge0: "\<And> i j. i < d \<Longrightarrow> j < d \<Longrightarrow> m ! i ! j \<succeq> 0"
    using mat0_index[of _ d _ d "0 :: 'a"] by auto
  note main = main[OF conjI[OF ge0 am] a1]
  obtain c where c: "c = ?n (fact ?d) * (a ^ ?d)" by auto
  have c0: "c \<succeq> 0"
    unfolding c
    by (rule mult_ge_zero[OF _ pow_ge_zero[OF a0]], auto)
  show ?thesis
  proof (rule exI, rule conjI[OF c0], intro allI impI)
    fix n i j :: nat
    assume i: "i < d" and j: "j < d" and n: "n > 0"
    let ?ij = "i - j"
    from i have ijd: "?ij \<le> ?d" by auto
    from main[rule_format, OF _ _ _ _ j i]
    have ge: "?n (fact (i - j)) * ((?n n * a) ^ (i - j)) \<succeq> mat_pow d m n ! i ! j" ..
    have na: "?n n * a \<succeq> 0" by (rule mult_ge_zero[OF _ a0], auto)
    have fact: "fact ?d \<ge> fact ?ij"
      by (rule fact_mono_nat[OF ijd])
    have nfact: "?n (fact ?d) \<succeq> ?n (fact ?ij)" 
      by (rule of_nat_mono[OF fact])
    have "?n n ^ ?d * c = ?n n ^ ?d * (?n (fact ?d) * a ^ ?d)"
      unfolding c by auto
    also have "... = ?n (fact ?d) * ((?n n) ^ ?d * a ^ ?d)" by (auto simp: field_simps)
    also have "((?n n) ^ ?d * a ^ ?d) = (?n n * a) ^ ?d"
      unfolding comm_semiring_1_class.normalizing_semiring_rules ..
    also have "?n (fact ?d) * (?n n * a) ^ ?d \<succeq> ?n (fact ?ij) * (?n n* a) ^ ?d"
      by (rule times_left_mono[OF pow_ge_zero[OF mult_ge_zero[OF _ a0]] nfact], auto)
    also have "... \<succeq> ?n (fact ?ij) * (?n n * a) ^ ?ij"
    proof (rule times_right_mono[OF of_nat_ge_zero pow_mono_exp[OF mult_ge_one[OF _ a1] ijd]]) 
      from n obtain nn where n: "n = Suc nn" by (cases n, auto)
      have "?n n \<succeq> 1 + 0" unfolding n of_nat_Suc
        by (rule plus_right_mono, auto)
      thus "?n n \<succeq> 1" by simp
    qed
    also have "... \<succeq> mat_pow d m n ! i ! j" using ge .
    finally
    show "c * (?n n ^ ?d) \<succeq> mat_pow d m n ! i ! j" by (simp add: field_simps)
  qed
qed 

lemma upper_triangular_mat_pow_value: assumes mat: "mat d d (m :: ('a :: bin_max_ordered_semiring_1) mat)"
  and tri: "upper_triangular m"
  and ge0: "mat_ge m (mat0 d d)"
  shows "\<exists> c. c \<succeq> 0 \<and> (\<forall> n > 0. ge (c * of_nat (n ^ (d - Suc 0))) (linear_norm (mat_pow d m n)))"
proof -
  from upper_triangular_mat_pow_index[OF mat tri ge0]
  obtain c where "c \<succeq> (0\<Colon>'a) \<and>
         (\<forall>n>0. \<forall>i<d. \<forall>j<d. c * of_nat n ^ (d - Suc 0) \<succeq>
                            mat_pow d m n ! i ! j)" ..
  hence c: "(c :: 'a) \<succeq> 0" and ge: "\<And> n i j. n > 0 \<Longrightarrow> i < d \<Longrightarrow> j < d \<Longrightarrow> c * of_nat n ^ (d - Suc 0) \<succeq> mat_pow d m n ! i ! j" by auto
  let ?c = "of_nat d * of_nat d * c"
  from c have c0: "?c \<succeq> 0" by auto
  show ?thesis
  proof (rule exI, rule conjI[OF c0], intro allI impI)
    fix n :: nat
    assume n: "0 < n"
    hence "?c * of_nat n ^ (d - Suc 0) \<succeq> linear_norm (mat_pow d m n)"
      using linear_norm_index[OF _ mat_pow[OF mat] ge[OF n]] c 
      by (auto simp: field_simps)
    thus "?c * of_nat (n ^ (d - Suc 0)) \<succeq> linear_norm (mat_pow d m n)"
      unfolding of_nat_power .
  qed
qed

context complexity_one_mono_ordered_semiring_1
begin
lemma upper_triangular_mat_pow_bound: assumes mat: "mat d d (m :: 'a mat)"
  and tri: "upper_triangular m"
  and ge0: "mat_ge m (mat0 d d)"
  shows "\<exists> c. (\<forall> n > 0. c * n ^ (d - Suc 0) \<ge> bound (linear_norm (mat_pow d m n)))"
proof -
  from upper_triangular_mat_pow_value[OF mat tri ge0]
  obtain c where c: "c \<succeq> 0" and ge: "\<And> n. n > 0 \<Longrightarrow> c * of_nat (n ^ (d - Suc 0)) \<succeq> linear_norm (mat_pow d m n)" by auto
  from bound_linear obtain c' where c': "\<And> n. bound (of_nat n) \<le> c' * n" by auto
  {
    fix n :: nat
    assume "n > 0"
    from bound_mono[OF ge[OF this]] have "bound (linear_norm (mat_pow d m n)) \<le> bound (c * of_nat (n ^ (d - Suc 0)))" by simp
    also have "... \<le> bound c * bound (of_nat (n ^ (d - Suc 0)))"
      by (rule bound_mult_of_nat)
    also have "... \<le> bound c * (c' * n ^ (d - Suc 0))"
      by (rule mult_le_mono2[OF c'])
    finally have "bound (linear_norm (mat_pow d m n)) \<le> (bound c * c') * n ^ (d - Suc 0)" by simp
  } note ge = this
  show ?thesis
    by (rule exI[of _ "bound c * c'"], insert ge, auto)
qed
end

locale matrix_carrier = SN_one_mono_ordered_semiring_1 d gt
  for gt :: "'a :: {show,max_ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: "'a"

locale mono_matrix_carrier = complexity_one_mono_ordered_semiring_1 gt d bound
  for gt :: "'a :: {show,bin_max_ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: 'a
  and bound :: "'a \<Rightarrow> nat" 
+ fixes mono :: "'a \<Rightarrow> bool"
  assumes mono: "\<And> x y z. mono x \<Longrightarrow> y \<succ> z \<Longrightarrow> x \<succeq> 0 \<Longrightarrow> x * y \<succ> x * z"

locale weak_complexity_linear_poly_order_carrier = 
  fixes weak_gt :: "'a :: {bin_max_ordered_semiring_1,show} \<Rightarrow> 'a \<Rightarrow> bool"
   and  default :: "'a"
   and  mono :: "'a \<Rightarrow> bool"
  assumes weak_gt_mono: "\<forall> x y. (x,y) \<in> set xys \<longrightarrow> weak_gt x y \<Longrightarrow> \<exists> gt bound. mono_matrix_carrier gt default bound mono \<and> (\<forall> x y. (x,y) \<in> set xys \<longrightarrow> gt x y)"
begin

abbreviation weak_mat_gt :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "weak_mat_gt \<equiv> mat_gtI weak_gt"

lemma weak_mat_gt_mono: assumes sd_n: "sd \<le> n" and
    orient: "\<And> m1 m2. mat n n m1 \<and> mat n n m2 \<and> (m1,m2) \<in> set m12s \<Longrightarrow> weak_mat_gt sd m1 m2"
   shows "\<exists> gt bound. mono_matrix_carrier gt default bound mono \<and> (\<forall> m1 m2. mat n n m1 \<and> mat n n m2 \<and> (m1, m2) \<in> set m12s \<longrightarrow> mat_gtI gt sd m1 m2)"
proof -
  let ?m1x = "concat (concat (map fst m12s))"
  let ?m2y = "concat (concat (map snd m12s))"
  let ?pairs = "concat (map (\<lambda> x. map (\<lambda> y. (x,y)) ?m2y) ?m1x)"
  let ?strict = "filter (\<lambda> (x,y). weak_gt x y) ?pairs"
  have "\<forall> x y. (x,y) \<in> set ?strict \<longrightarrow> weak_gt x y" by auto
  from weak_gt_mono[OF this] obtain gt bound where order: "mono_matrix_carrier gt default bound mono" and orient2: "\<And> x y. (x, y) \<in> set ?strict \<Longrightarrow> gt x y" by auto
  show ?thesis
  proof (intro exI allI conjI impI, rule order)
    fix m1 m2
    assume ass: "mat n n m1 \<and> mat n n m2 \<and> (m1,m2) \<in> set m12s"
    hence wf1: "mat n n m1" and wf2: "mat n n m2" and m1m2: "(m1,m2) \<in> set m12s" by auto
    from orient[OF ass, unfolded mat_gt_index[OF wf1 wf2 sd_n]] 
    obtain i j where ge: "mat_ge m1 m2" and i: "i < sd" and j: "j < sd" and weak_gt: "weak_gt (m1 ! i ! j) (m2 ! i ! j)" (is "weak_gt ?x ?y") by auto
    from wf1 i j sd_n have m1ij: "m1 ! i \<in> set m1 \<and> m1 ! i ! j \<in> set (m1 ! i)" unfolding mat_def by (auto simp: vec_def list_all_iff)
    from wf2 i j sd_n have m2ij: "m2 ! i \<in> set m2 \<and> m2 ! i ! j \<in> set (m2 ! i)" unfolding mat_def  by (auto simp: vec_def list_all_iff)
    have x: "?x \<in> set ?m1x" by (auto, rule bexI[OF _ m1m2], rule bexI[of _ "m1 ! i"], auto simp: m1ij)
    have y: "?y \<in> set ?m2y" by (auto, rule bexI[OF _ m1m2], rule bexI[of _ "m2 ! i"], auto simp: m2ij)
    from x y have "(?x,?y) \<in> set ?pairs" by force
    with weak_gt have gt: "(?x,?y) \<in> set ?strict" by simp
    show "mat_gtI gt sd m1 m2" unfolding mat_gt_index[OF wf1 wf2 sd_n]
      by (intro exI conjI, rule ge, rule i, rule j, rule orient2[OF gt])
  qed
qed
end

sublocale mono_matrix_carrier \<subseteq> SN_strict_mono_ordered_semiring_1 d gt mono
proof
  show "SN {(x,y). y \<succeq> 0 \<and> x \<succ> y}" 
    unfolding SN_def
    by (intro allI, rule deriv_bound_SN_on, rule bound)
next
  fix x y z :: 'a
  assume "mono x" "y \<succ> z" "x \<succeq> 0"
  thus "x * y \<succ> x * z"
    by (rule mono)
qed

sublocale mono_matrix_carrier \<subseteq> matrix_carrier ..



end