(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Min_Set_Order
  imports List_Order
begin

definition min_set_s_order :: "'a list_ext" where "min_set_s_order s ns \<equiv> {(as,bs). bs \<noteq> [] \<and> (\<forall> a \<in> set as. \<exists> b \<in> set bs. (a,b) \<in> s)}"

definition min_set_ns_order :: "'a list_ext" where "min_set_ns_order s ns \<equiv> {(as,bs). (\<forall> a \<in> set as. \<exists> b \<in> set bs. (a,b) \<in> ns)}"

fun helper where "helper f S 0 = (SOME a . a \<in> set (f (Suc 0)))"
  | "helper f S (Suc n) = (SOME b. (helper f S n, b) \<in> S \<and> b \<in> set (f (Suc (Suc n))))"

interpretation list_order_extension min_set_s_order min_set_ns_order
proof(intro list_order_extension.intro)
  fix s ns
  assume op: "SN_order_pair s ns"
  then interpret SN_order_pair s ns .
  let ?S = "min_set_s_order s ns"
  let ?NS = "min_set_ns_order s ns"
  note d = min_set_s_order_def min_set_ns_order_def
  show "SN_order_pair ?S ?NS"
  proof 
    show "refl ?NS" using refl_NS unfolding refl_on_def d by auto
  next
    show "trans ?S" using trans_S unfolding trans_def d by blast
  next
    show "trans ?NS" using trans_NS unfolding trans_def d by blast
  next
    show "?NS O ?S \<subseteq> ?S" using compat_NS_S_point unfolding d by blast
  next
    show "?S O ?NS \<subseteq> ?S" using compat_S_NS_point unfolding d by force
  next
    show "SN ?S"
    proof
      fix f
      assume "\<forall> i. (f i, f (Suc i)) \<in> ?S"
      hence f: "\<And> i. (f i, f (Suc i)) \<in> ?S" by auto
      {
        fix i        
        have "\<exists> x. x \<in> set (f (Suc i))" using f[of i] unfolding d by (cases "f (Suc i)", auto)
      } note ne = this
      obtain h where h: "h = helper f s" by auto
      {
        fix n
        have "(h n, h (Suc n)) \<in> s \<and> h (Suc n) \<in> set (f (Suc (Suc n)))"
        proof (induct n)
          case 0
          from someI_ex[OF ne[of 0]] have "h 0 \<in> set (f (Suc 0))" unfolding h by auto
          with f[of "Suc 0"] have "\<exists> x. (h 0, x) \<in> s \<and> x \<in> set (f (Suc (Suc 0)))"
            unfolding d by auto
          from someI_ex[OF this] show ?case unfolding h by auto 
        next
          case (Suc n)
          with f[of "Suc (Suc n)"] have "\<exists> x. (h (Suc n), x) \<in> s \<and> x \<in> set (f (Suc (Suc (Suc n))))" unfolding d by auto
          from someI_ex[OF this] show ?case unfolding h by auto
        qed
      }
      with SN show False by auto
    qed
  qed
next
  fix S f NS as bs
  assume "\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a,f b) \<in> S" 
    "(as,bs) \<in> min_set_s_order S NS"
  thus "(map f as, map f bs) \<in> min_set_s_order S NS" unfolding min_set_s_order_def by force
next
  fix S f NS as bs
  assume "\<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a,f b) \<in> NS" 
    "(as,bs) \<in> min_set_ns_order S NS"
  thus "(map f as, map f bs) \<in> min_set_ns_order S NS" unfolding min_set_ns_order_def by force
next
  fix as bs :: "'a list" and NS S
  assume "length as = length bs" "\<And> i. i < length bs \<Longrightarrow> (as ! i, bs ! i) \<in> NS"
  thus "(as,bs) \<in> min_set_ns_order S NS" unfolding min_set_ns_order_def set_conv_nth by force
qed

hide_const helper
end
