(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Rewriting_Impl
imports 
  Innermost_Usable_Rules_Impl
  Critical_Pairs_Impl
  Rewriting
begin


(* TODO: one might automatically put st' into strict pairs,
   if st is not strict and there are strict rules *)
definition
  rewriting_proc ::
    "('dpp, 'f :: {linorder,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> pos \<Rightarrow> 'dpp proc"
where
  "rewriting_proc I U st st' lr p dpp \<equiv> 
    check_return (do {
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');
     let R = dpp_ops.rules I dpp;
     let s = fst st;
     let t = snd st;
     let t' = snd st';
     check (s = fst st') (shows ''left-hand sides of old and new pair differ'');
     check_rstep' R p lr t t';
     check_subseteq U R <+? (\<lambda> lr. shows_rule lr +@+ shows '' is not a rule of the rewrite system '');
     check (st \<in> set (dpp_ops.P I dpp) \<or> dpp_ops.R I dpp = []) (shows ''strict DP or no strict rules required'');
     let tp = t |_ p;
     let urc = is_ur_closed_impl_dpp_mv I dpp U;
     let check_urc = (\<lambda> S t. check (urc S t) (shows ''term '' +@+ shows t +@+ shows '' is not closed under usable rules''));
     check_subseteq (vars_term_list (snd lr)) (vars_term_list (fst lr)) <+? (\<lambda> x. shows ''variable condition in rule violated'');
     check_subseteq (vars_term_list tp) (vars_term_list s) <+? (\<lambda> x. shows ''variable condition in pair violated'');
     check_urc [s] tp;
     check_allm (\<lambda>(l,r). check_urc (args l) r) U;
     check_critical_pairs_innermost U <+? (\<lambda> s. shows ''problem in showing UNF of usable rules'' +@+ shows_nl +@+ s);
     check_allm (\<lambda>(b,s,t). check (s = t) (shows ''non-trivial critical pair between rule to rewrite and usable rules'')) (critical_pairs_impl [lr] U)
   })
   (dpp_ops.replace_pair I dpp st [st'])"

lemma rewriting_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (rewriting_proc I (U :: ('f :: {linorder,show},string)rules)
  st st' lr p)"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "rewriting_proc I U st st' lr p d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?U = "set U"
    let ?Rb = "?R \<union> ?Rw"
    let ?Rb' = "set (rules d)" 
    obtain s t where st: "st = (s,t)" by force
    obtain l r where lr: "lr = (l,r)" by force
    obtain s' t' where st': "st' = (s',t')" by force
    note ok = ok[unfolded rewriting_proc_def st lr st' Let_def dpp_spec_sound is_ur_closed_impl_dpp_mv[OF I], simplified] 
    from ok have NF: "NF_terms ?Q \<subseteq> NF_trs ?Rb" and NF': "NF_terms ?Q \<subseteq> NF_trs ?Rb'" by auto
    from ok have d': "d' = replace_pair d (s,t) [(s,t')]" by auto
    have d': "dpp d' = (replace (s,t) {(s,t')} ?P, replace (s,t) {(s,t')} ?Pw, ?Q, ?R, ?Rw)" 
      unfolding d' unfolding replace_pair_sound by simp
    note fin = fin[unfolded this]
    interpret R_Q_U_ecap ?Rb' ?U ?Q icap'
      by (unfold_locales, rule icap, rule NF')
    have Rb: "?Rb = ?Rb'" by simp
    from ok obtain \<sigma> where step: "(t,t') \<in> rstep_r_p_s ?Rb' (l,r) p \<sigma>" unfolding Rb by blast
    have subset: "?U \<subseteq> ?Rb'" using ok by simp
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
    proof (rule finite_under_var_cond)
      assume var_cond: "\<And> lr. lr \<in> ?Rb \<Longrightarrow> is_Fun (fst lr)"
      show "finite_dpp (?P,?Pw,?Q,?R,?Rw)"
      proof (rule rewriting_proc[OF fin step _ _ _ _ _ _ subset _ Rb])
        fix l r
        assume "(l,r) \<in> ?U"
        with ok
        show "is_ur_closed_term_mv full_af (set (args l)) r" by auto
      next
        show "(s,t) \<in> ?P \<or> ?R = {}" using ok by auto
      next
        show "vars_term r \<subseteq> vars_term l" using ok by auto
      next
        show "vars_term (t |_ p) \<subseteq> vars_term s" using ok by auto
      next
        from ok 
        show "is_ur_closed_term_mv full_af {s} (t |_ p)" by auto
      next
        from ok have ok: "isOK(check_critical_pairs_innermost U)" by simp
        from NF_Q_R NF_anti_mono[OF rstep_mono[OF subset]] have NFQU: "NFQ \<subseteq> NF_trs (set U)" by auto
        {
          fix l r
          assume "(l,r) \<in> ?U"
          with subset Rb[symmetric] var_cond[of "(l,r)"] have "is_Fun l" by auto
        }
        from check_critical_pairs_innermost[OF ok NFQU this]
        have "CR (qrstep ?Q ?U)" .
        from CR_imp_UNF[OF this] show "UNF (qrstep ?Q ?U)" .
      next
        fix b s t
        assume "(b,s,t) \<in> critical_pairs {(l,r)} (set U)"
        with ok show "s = t" by auto
      qed
    qed
  qed
qed

end