(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Subterm_Criterion_Impl
imports
  Subterm_Criterion
  QDP_Framework_Impl
  Q_Restricted_Rewriting_Impl
begin

definition
  check_strict_rstep ::
    "('f::show, 'v::show) rules \<Rightarrow>
     (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option) \<Rightarrow>
     'f proj \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows check"
where
  "check_strict_rstep R rseqm p r = (
     let s = proj_term p (fst r); t = proj_term p (snd r) in
     case rseqm r of
       None \<Rightarrow> check_supt s t
     | Some rseq \<Rightarrow> (if length rseq = 0
       then check_supt s t
       else (
         let u = snd (snd (last rseq)) in (
           check_qrsteps (\<lambda> _. True) R rseq s u
         \<guillemotright> check_supteq u t))))"

definition
  check_strict_qrstep ::
    "(('f::show, 'v::show) term \<Rightarrow> bool) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option) \<Rightarrow>
     'f proj \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows check"
where
  "check_strict_qrstep nf R rseqm p r = (
    let s = proj_term p (fst r); t = proj_term p (snd r) in
    case rseqm r of
      None \<Rightarrow> check_supt s t
    | Some [(pos, rule, u)] \<Rightarrow> (
        check_qrstep nf R pos rule s u
      \<guillemotright> check_supteq u t)
    | Some _ \<Rightarrow> error (shows_string ''more than a single Q-rewrite step is not allowed''))"

lemma check_strict_rstep_sound:
  assumes "isOK (check_strict_rstep R rseqm p r)"
  shows "(proj_term p (fst r), proj_term p (snd r)) \<in> (op \<rhd> \<union> rstep (set R))^+"
    (is "(?s, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+")
proof (cases "rseqm r")
  case None
  from assms[unfolded check_strict_rstep_def None]
    have "isOK (check_supt ?s ?t)" by simp
  hence "(?s, ?t) \<in> op \<rhd>" by simp
  thus ?thesis by auto
next
  case (Some rseq)
  show ?thesis
  proof (cases "length rseq = 0")
    case True
    with assms[unfolded check_strict_rstep_def Some]
      have "isOK (check_supt ?s ?t)" by simp
    hence "(?s, ?t) \<in> op \<rhd>" by simp
    thus ?thesis by auto
  next
    let ?u = "snd (snd (last rseq))"
    case False
    with assms[unfolded check_strict_rstep_def Some]
      have "isOK (check_qrsteps (\<lambda> _. True) R rseq ?s ?u \<guillemotright> check_supteq ?u ?t)" unfolding Let_def by simp
    hence "(?s, ?u) \<in> (rstep (set R))^^(length rseq)" and "(?u, ?t) \<in> op \<unrhd>" 
      using check_qrsteps_sound[of "\<lambda> _. True" "[]" R rseq ?s ?u] by auto
    with False obtain n where "(?s, ?u) \<in> (rstep (set R))^^(Suc n)" by (induct rseq) simp_all
    hence "(?s, ?u) \<in> (rstep (set R))^+" using trancl_power[of _ "rstep (set R)"] by blast
    from `(?u, ?t) \<in> op \<unrhd>` show ?thesis unfolding supteq_supt_conv
    proof
      assume "?u = ?t"
      from `(?s, ?u) \<in> (rstep (set R))^+`
        show ?thesis unfolding `?u = ?t` using trancl_union_right[of "rstep (set R)" "op \<rhd>"] by blast
    next
      assume "(?u, ?t) \<in> op \<rhd>"
      hence "(?u, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+" by auto
      moreover from `(?s, ?u) \<in> (rstep (set R))^+`
        have "(?s, ?u) \<in> (op \<rhd> \<union> rstep (set R))^+" using trancl_union_right by blast
      ultimately show ?thesis by simp
    qed
  qed
qed

lemma check_strict_qrstep_sound:
  assumes nf: "\<And>t. nf t \<longleftrightarrow> t \<in> NF_terms (set Q)"
  assumes ok: "isOK (check_strict_qrstep nf R rseqm p r)"
  shows "(proj_term p (fst r), proj_term p (snd r)) \<in> (op \<rhd> \<union> qrstep (set Q) (set R) O op \<unrhd>)"
    (is "(?s, ?t) \<in> (op \<rhd> \<union> ?S)")
proof (cases "rseqm r")
  case None
  from ok[unfolded check_strict_qrstep_def None]
    show ?thesis by auto
next
  case (Some prus)
  show ?thesis
  proof (cases prus)
    case Nil
    from ok[unfolded check_strict_qrstep_def Some Nil]
      show ?thesis by simp
  next
    case (Cons pru prus')
    note Cons' = this
    show ?thesis
    proof (cases prus')
      case (Cons pru' prus'')
      from ok[unfolded check_strict_qrstep_def Some Cons' Cons]
        show ?thesis by simp
    next
      case Nil
      with Cons obtain p r u where pru: "pru = (p, r, u)" by (cases pru) auto
      with Nil and Cons have prus: "prus = [(p, r, u)]" by simp
      from nf and ok[unfolded check_strict_qrstep_def Some Cons Nil, simplified]
        have okstep: "isOK (check_qrstep nf R p r ?s u)"
        and supteq: "isOK (check_supteq u ?t)" unfolding pru by auto
      from check_qrstep_qrstep[OF nf okstep]
        and supteq[unfolded isOK_check_supteq]
        show ?thesis by blast
    qed
  qed
qed

definition
  check_weak :: "'f proj \<Rightarrow> ('f::show, 'v::show) rule \<Rightarrow> shows check"
where
  "check_weak p r \<equiv> check (proj_term p (fst r) = proj_term p (snd r))
     (''the projected lhs is not equal to the projected rhs'' +#+ shows_nl)
   <+? (\<lambda>s. ''Could not orient rule '' +#+ shows_rule r +@+ '', since'' +#+ shows_nl +@+
          shows (proj_term p (fst r)) +@+ '' != '' +#+ shows (proj_term p (snd r)) +@+
          shows_nl +@+ s)"

lemma check_weak_sound[simp]:
  assumes "isOK (check_weak p r)"
  shows "proj_term p (fst r) = proj_term p (snd r)"
  using assms unfolding check_weak_def by auto


definition
  subterm_criterion_proc ::
    "('dpp, 'f::show, 'v::show) dpp_ops \<Rightarrow>
    'f proj \<Rightarrow> (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option) \<Rightarrow>
    ('f, 'v) rules \<Rightarrow> 'dpp proc"
where
  "subterm_criterion_proc I p rseqm Prm dpp = check_return (do {
    let P       = dpp_ops.pairs I dpp;
    let R       = dpp_ops.rules I dpp;
    let P'      = snd (dpp_ops.split_pairs I dpp Prm);
    check_allm (\<lambda>(l, r). do {
      check_no_var l;
      check_no_var r;
      check_no_defined_root (dpp_spec.is_defined I dpp) r
    }) P;
    (if dpp_ops.Q_empty I dpp
      then check_allm (check_strict_rstep R rseqm p) Prm
      else check_allm (check_strict_qrstep (dpp_ops.is_QNF I dpp) R rseqm p) Prm);
    check_allm (check_weak p) P'
  }) (dpp_spec.delete_pairs I dpp Prm)"

context dpp_spec
begin

lemma subterm_criterion_proc:
  "sound_proc_impl (subterm_criterion_proc I p rseqm ps)"
proof
  fix d d'
  assume ok: "subterm_criterion_proc I p rseqm ps d = return d'"
    and fin: "finite_dpp (dpp_ops.dpp I d')"
  let ?S  = "set ps"
  let ?P  = "set (dpp_ops.P I d)"
  let ?Pw = "set (dpp_ops.Pw I d)"
  let ?P' = "?P \<union> ?Pw"
  let ?Q  = "set (dpp_ops.Q I d)"
  let ?R  = "set (dpp_ops.R I d)"
  let ?Rw = "set (dpp_ops.Rw I d)"
  let ?R' = "?R \<union> ?Rw"
  let ?\<pi>  = "proj_term p"
  show "finite_dpp (dpp_ops.dpp I d)"
  proof (cases "dpp_ops.split_pairs I d ps")
    case (Pair D P')
    from split_pairs_sound[OF Pair]
      have D: "set D = (?P \<union> ?Pw) \<inter> ?S"
      and P': "set P' = (?P \<union> ?Pw) - ?S" by (auto simp: dpp_spec_sound)
    let ?D = "set D"
    note ok = ok[unfolded subterm_criterion_proc_def, simplified, simplified Pair]
    from ok
      have cond: "\<forall>(l, r)\<in>?P'.
        is_Fun l \<and> is_Fun r \<and> \<not> defined ?R' (the (root r))"
      by (simp add: split_def Let_def
          check_no_defined_root_sound[OF is_defined_sound, unfolded rules_map_defined_conv]
          rules_sound pairs_sound)
    from ok
      have weak: "\<forall>(l, r)\<in>set P'. ?\<pi> l = ?\<pi> r"
      by (simp_all add: split_def Let_def pairs_sound Pair)
    from ok
      have strict: "?Q = {} \<and> (\<forall>(l, r)\<in>?S. (?\<pi> l, ?\<pi> r) \<in> (op \<rhd> \<union> rstep ?R')^+) \<or>
                              (\<forall>(l, r)\<in>?S. (?\<pi> l, ?\<pi> r) \<in> op \<rhd> \<union> qrstep ?Q ?R' O op \<unrhd>)"
      using check_strict_rstep_sound[of "dpp_ops.rules I d" rseqm p]
      using check_strict_qrstep_sound[OF is_QNF_sound[THEN fun_cong],
        of d "dpp_ops.rules I d" rseqm p]
      by (cases "?Q = {}")
         (simp_all add: split_def Pair rules_sound Q_empty_sound D P'
                   del: is_QNF_sound)
    from ok have d': "d' = dpp_spec.delete_pairs I d ps" by simp
    have proc: "Subterm_Criterion.subterm_criterion_proc p ?S
      (?P, ?Pw, ?Q, ?R, ?Rw) (?P - ?S, ?Pw - ?S, ?Q, ?R, ?Rw)"
      by (simp add: subterm_criterion_cond_def cond P' weak)
         (insert strict, auto) 
    note fin = fin[unfolded d' delete_P_Pw_sound]
    show ?thesis unfolding dpp_spec_sound d'
      by (rule ci_proc[OF fin proc ci_subterm_criterion_proc])
  qed
qed

end

datatype 'f projL = Projection "(('f \<times> nat) \<times> nat) list"

fun
  create_proj :: "'f projL \<Rightarrow> 'f proj"
where
  "create_proj (Projection p) f = (case lookup f p of None \<Rightarrow> 0 | Some n \<Rightarrow> n)"

fun
  create_rseq_map :: "(('f, 'v) rule \<times> ('f, 'v) rseq) list \<Rightarrow> (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option)"
where
  "create_rseq_map rseqs r = lookup r rseqs"

end
