(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Termination_Switch_Impl
imports
  Innermost_Switch_Impl
  Termination_Switch
  Innermost_Impl
begin

definition
  switch_termination_proc :: "('dpp, 'f::show, string) dpp_ops \<Rightarrow> 'f join_info \<Rightarrow> 'dpp proc"
where
  "switch_termination_proc I joins_i dpp \<equiv> let 
       P = dpp_ops.pairs I dpp;
       R = dpp_ops.rules I dpp;
       Q = dpp_ops.Q I dpp
     in 
  check_return (do {
     let cp = critical_pairs_impl R R;
     check_allm (\<lambda> (b,st). check b (shows ''rules are not overlay'')) cp;
     check_critical_pairs R cp joins_i;
     check (critical_pairs_impl P R = []) (shows ''there are overlaps between P and R'');
     check_NF_terms_subset (map fst R) Q
          <+? (\<lambda> q. shows q +@+ shows '' is not in normal form w.r.t. R'')
   }) (dpp_ops.mk I P [] [] [] R)"


lemma switch_termination_proc: fixes I :: "('a,'f :: show,string)dpp_ops"
  assumes I: "dpp_spec I" 
  and inf: "infinite (UNIV :: 'f set)"
  and ok: "switch_termination_proc I joins_i dpp = return dpp'"
  and infinite: "infinite_dpp (set (dpp_ops.pairs I dpp'), set (dpp_ops.Q I dpp'), set (dpp_ops.rules I dpp'))"
  shows "infinite_dpp (set (dpp_ops.pairs I dpp), set (dpp_ops.Q I dpp), set (dpp_ops.rules I dpp))"
proof -
  interpret dpp_spec I by (rule I)
  let ?P = "set (pairs dpp)"
  let ?R = "set (rules dpp)"
  let ?Q = "set (Q dpp)"
  note ok = ok[unfolded switch_termination_proc_def Let_def, simplified]
  from ok infinite have infinite: "infinite_dpp (?P, {}, ?R)"  by simp
  from ok critical_pairs_impl[of "pairs dpp" "rules dpp"] 
    have cps_empty: "critical_pairs ?P ?R = {}" by simp
  from ok check_critical_pairs[of "rules dpp"] have WCR: "WCR (rstep ?R)" by auto
  from ok have NF: "NF_trs ?R \<subseteq> NF_terms ?Q" by auto
  from WCR have WCR_SN: "WCR_on (rstep ?R) {t. SN_on (rstep ?R) {t}}"
    unfolding WCR_on_def by auto
  from ok have overlay: "\<And> l r. (False,l,r) \<notin> critical_pairs ?R ?R" by auto
  note innermost_switch = switch_to_innermost_locally_confluent_overlay_finite[OF WCR_SN overlay _ finite_set inf]
  have SN: "(SN (qrstep ?Q ?R)) \<Longrightarrow> SN (rstep ?R)" 
  proof (rule innermost_switch)
    assume SN: "SN (qrstep ?Q ?R)"
    show "SN (qrstep (lhss ?R) ?R)"
      by (rule SN_subset[OF SN qrstep_mono], insert NF, auto)
  qed
  show ?thesis
    by (rule termination_switch_proc[OF infinite cps_empty WCR NF SN])
qed


end